# -*- coding: utf-8 -*-
"""
    Module: aypylib : msoffice
    File:   ppt_rep.py
    Descr.: Python-pptx helper class: simple ppt report
    Update: 190619 v0.0 build for pyPTself
            191006 v0.1 mod for helper class
            210725 v1.0
    Last:   221228 v1.1 mod wrText()
@author: Ayoshida(RIKEN,Nishina center)
"""

""" import: standard lib.
----.---.---.---.---.---.---.---.---.---.---.---.---"""
from    PIL  import Image
import  io  # use BytesIO()
        
""" import: python-pptx  package  (used v0.6.18)
    https://python-pptx.readthedocs.io/en/latest/
    > pip install python-pptx
----.---.---.---.---.---.---.---.---.---.---.---.---"""
from pptx             import Presentation
from pptx.util        import Inches, Mm, Pt # Inches, Cm, 
#from pptx.enum.shapes import MSO_SHAPE
#from pptx.enum.dml    import MSO_COLOR_TYPE, MSO_THEME_COLOR
from pptx.enum.text   import MSO_ANCHOR, MSO_AUTO_SIZE, PP_ALIGN
from pptx.dml.color   import RGBColor

""" import: my custom lib.       
----.---.---.---.---.---.---.---.---.---.---.---.---"""
#import  aypylib             as ay 
#from    aypylib.plot        import  (MPlot, Hist1D)

""" import from this module
----.---.---.---.---.---.---.---.---.---.---.---.---"""

""" .===.===.===.===.===.===.===.===.===.===.===.===.===.===.===.===
        Class   PptRep      simple ppt report
====.===.===.===.===.===.===.===.===.===.===.===.===.===.===.=== === """
class pptRep(object):
    """ Python-pptx helper class : simple ppt report
Params. for constructor:
    paper:  str     Paper size eg) 'A2','A3','A4'(defa))
Attributes:
    see ._myclear()
----.---.---.---.---.---.---.---.---.---.---.---.---.---"""

    """ class Attributes
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    def _clear_pptRep(self): 
        self.ppt    = None              # ppt Presentation
        self.lay    = None              # ppt layout
        self.PgHmm,self.PgWmm = 0,0     # [mm] 用紙サイズ 横置き
        """ defa. settings """
        self.fntNm  = 'ＭＳ ゴシック'       # =等幅  'ＭＳ Ｐゴシック' は等幅でない
        self.fntSiz = Pt(12)
        self.fntCol = 'k'               # black
        
    def __init__(self, paper:str='A4',ori:str='land'):
        """ Constructor
        <optional>
        paper:  str     Paper size eg) 'A2','A3','A4'(defa))
        ori:    str     Paper orientation 'port'縦置き / 'land'横置き=defa
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        self._clear_pptRep()
        self.ppt = Presentation()
        self.setPaper(paper,ori)      
        self.lay = self.ppt.slide_layouts[6]  # レイアウト, 6番は白紙
        # ppt.slides.add_slide(ppLyo)         # 白紙のスライドの追加
        
    def setPaper(self, psz:str, ori:str='landscape'):   
        """  set paper size     
        psz:    str     Paper size eg) 'A2','A3','A4'
        ori:    str     Orientation 'portrait'縦置き / 'landscape'横置き=defa
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        h,w = ppt_paper_size(psz)
        self.PgHmm, self.PgWmm = h,w
        h,w = (w,h) if('port' in ori) else (h,w)
        self.ppt.slide_height = Mm(h)
        self.ppt.slide_width  = Mm(w)
            
    def addSlide(self):
        """ add new slide
        <Ret.>
        sld:    Slide
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        sld = self.ppt.slides.add_slide(self.lay)
        return( sld )

    def wrText(self, sld, txt:str, lt:tuple, *,
                   fs:int=None, fc:str=None, fn:str=None,
                   it:bool=False, bd:bool=False, ul:bool=False):
        """ write text box on sld
        sld:    Slide
        txt:    str          text to write
        lt:  (left,top) [mm] Left,Top corner pos
        <optional>
        fs:     int     [Pt] font size, if None use self.fntSiz
        fc:     str          font color eg) 'w','r', ...
        fn:     str          font name, if None use self.fntNm
        it:     bool         Italic
        bd:     bool         Bold
        ul:     bool         Under Line
        <color> fc R:G:B
        'c': (0.0, 0.75, 0.75), 'b': (0.0, 0.0, 1.0),
        'w': (1.0, 1.0, 1.0),   'g': (0.0, 0.5, 0.0), 
        'y': (0.75, 0.75, 0),   'k': (0.0, 0.0, 0.0), 
        'r': (1.0, 0.0, 0.0),   'm': (0.75, 0, 0.75)}
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        left,top = lt
        txb = sld.shapes.add_textbox(left=Mm(left),top=Mm(top),\
                        width=Mm(10),height=Mm(5))
        txf = txb.text_frame
        txf.clear()
        prg = txf.paragraphs[0]
        prg.text = txt
        col = self.fntCol if fc is None else fc
        siz = self.fntSiz if fs is None else Pt(fs)
        fnm = self.fntNm  if fn is None else fn
        prg.font.color.rgb = ppt_color(col)
        prg.font.size   = siz
        prg.font.name   = fnm
        prg.font.italic = it
        prg.font.bold   = bd
        prg.font.underline = ul
        self.set_txf(txf)

    def wrImage(self, sld, im:Image, lt:tuple, *,
                wid:int=None, hgt:int=None):
        """ write PIL Image on sld    -> r#04)
        im:     PIL.Image
        lt:  (left,top) [mm] Left,Top corner pos
        <optional>
        wid:    int     [mm] width
        hgt:    int     [mm] height
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        b = io.BytesIO()
        im.save(b,'PNG')
        b.seek(0)
        self.wrIoByte(sld,b,lt,wid=wid,hgt=hgt)

    def wrFigure(self, sld, fig:object, lt:tuple, *,
                wid:int=None, hgt:int=None):
        """ write matplot.figure on sld
        fig:    matplotlib.Figure 
        lt:  (left,top) [mm] Left,Top corner pos
        <optional>
        wid:    int     [mm] width
        hgt:    int     [mm] height
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        b = io.BytesIO()
        fig.savefig(b,format='PNG')
        b.seek(0)
        self.wrIoByte(sld,b,lt,wid=wid,hgt=hgt)

    def wrIoByte(self, sld, ios, lt:tuple, *,
                wid:int=None, hgt:int=None):
        """ write io.BytesIO on sld    -> r#05)
        ios:    io.BytesIO   Byte stream IO
        lt:  (left,top) [mm] Left,Top corner pos
        <optional>
        wid:    int     [mm] width
        hgt:    int     [mm] height
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        left,top = lt
        w = Mm(wid) if(wid is not None) else None
        h = Mm(hgt) if(hgt is not None) else None      
        ios.seek(0)
        shp = sld.shapes
        shp.add_picture(ios,left=Mm(left),top=Mm(top),
            width=w,height=h)

    def save_ppt(self, fn):
        """ save Presentation to ppt file
        fn:     str     output file name eg)'a.ppt'
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        self.ppt.save(fn)
 
    def set_pgp(self, pgp, fntNm, fntSiz, fntCol, bld ):   
        """  set paragraph
        pgp:    Paragraph
        fntNm:  str         font name eg) 'Calibri'
        fntCol: RGBColor(r,g,b)
        bld:    bool        bold
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        pgp.font.name   = fntNm
        pgp.font.size   = fntSiz
        pgp.font.color.rbg  = fntCol
        pgp.font.bold   = bld
        pgp.line_spacing= 0.7
        pgp.alignment   = PP_ALIGN.LEFT
        
    def set_txf(self, txf ):
        """  set TextFrame Margine = tight
        txf:    TextFrame
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        txf.margin_left   = Mm(0)
        txf.margin_top    = Mm(0)
        txf.margin_right  = Mm(0)
        txf.margin_bottom = Mm(0)
        txf.vertical_anchor = MSO_ANCHOR.TOP
        txf.word_wrap = False
        txf.auto_size = MSO_AUTO_SIZE.SHAPE_TO_FIT_TEXT
        
    def add_txrun(self, pgp, txt, fntNm, fntSiz, fntCol, bld):
        """  xt 途中で属性を変える時は runを追加
        pgp:    Paragraph
        fntNm:  str         font name eg) 'Calibri'
        fntCol: RGBColor(r,g,b)
        bld:    bool        bold     
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        run = pgp.add_run()
        run.text        = txt
        run.font.name   = fntNm
        run.font.size   = fntSiz
        run.font.color.rgb = fntCol
        run.font.bold   = bld
        run.font.italic = None

    
""" .===.===.===.===.===.===.===.===.===.===.===.===.===.===.===.===
        Helper Funcs
====.===.===.===.===.===.===.===.===.===.===.===.===.===.===.=== === """
def ppt_paper_size( paper:str ) -> tuple:
    """  ISO paper size in mm
    paper:      str     Paper size eg) 'A4','B3', ...
    <Ret.>
    (hmm,wmm):  tuple.int   [mm] landscape:横置き hight, width
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    psiz = {    # A,B版（JIS規格サイズ） [mm]
        'A0':( 841,1189),'A1':(594, 841),'A2':(420,594),
        'A3':( 297, 420),'A4':(210, 297),'A5':(148, 210),
        'A6':( 105, 148),
        'B0':(1030,1456),'B1':(728,1030),'B2':(515,728),
        'B3':( 364, 515),'B4':(257, 364),'B5':(182, 257),
        'B6':( 128, 182) }
    if(paper not in psiz):
        raise ValueError('paper={} not found'.format(paper))
    hmm,wmm = psiz[paper]
    return( (hmm,wmm) )

def ppt_color( col:str ):
    """  color name -> RGBColor(r,g,b)
    col:      str     Color eg) 'w','r', ...
    <Ret.>
    RGBColor(r,g,b):  for python-pptx
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    colors = {      # -> r#03)
        'c': (0.0, 0.75, 0.75), 'b': (0.0, 0.0, 1.0),
        'w': (1.0, 1.0, 1.0),   'g': (0.0, 0.5, 0.0), 
        'y': (0.75, 0.75, 0),   'k': (0.0, 0.0, 0.0), 
        'r': (1.0, 0.0, 0.0),   'm': (0.75, 0, 0.75)}
    if(col not in colors):
        raise ValueError('col={} not found'.format(col))
    r,g,b = colors[col]
    return( RGBColor(int(r*255),int(g*255),int(b*255)) )

  
""" ========================================================
            Test Main
    ========================================================"""
if __name__ == '__main__':

    """ Test-CTRL Sw ---.---.---.---.---.---.---.---.---"""
    SwTest11 =  True    # test 
    """-.---.---.---.---.---.---.---.---.---.---.---.---"""
    rep = pptRep('A4')
    
    """ test11 : rep_scan
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    if(SwTest11):
        sld = rep.addSlide()
        rep.wrText(sld,'sample1',(20,20),fs=20,fc='r')
        rep.wrText(sld,'sample2',(50,20),fs=30,fc='b',fn='ＭＳ Ｐ明朝')

    rep.save_ppt('_a.pptx')
        
    print('@@main: -- Fin ---')     # 以上です。

""" ========================================================
        ref) Numbered ref
    ========================================================
r#03) matplotlib.colors
     https://het.as.utexas.edu/HET/Software/Matplotlib/api/colors_api.html

cache = {
    u'blue': (0.0, 0.0, 1.0),
    u'black': (0.0, 0.0, 0.0),
    u'white': (1.0, 1.0, 1.0),
    u'gray': (0.50,0.50,0.50),
    u'c': (0.0, 0.75, 0.75), 
    u'b': (0.0, 0.0, 1.0), 
    u'g': (0.0, 0.5, 0.0), 
    u'k': (0.0, 0.0, 0.0), 
    u'm': (0.75, 0, 0.75), 
    u'r': (1.0, 0.0, 0.0),
    u'w': (1.0, 1.0, 1.0), 
    u'y': (0.75, 0.75, 0) }
      
r#04) stack ovf convert PIL Image object to File object
    https://stackoverflow.com/questions/46593477/convert-pil-image-object-to-file-object/51553992
    
r#05) Quiita matplotlibのグラフをローカルに一時保存せずにslackへ送信するサンプルコード
    https://qiita.com/kaito__/items/9aa63cccc99261814065
    
    ========================================================
        ref) Others
    ======================================================== 
・Qiita pythonで作ったパワーポイントで自己紹介してみた
     https://qiita.com/Ishio/items/54912dc3ead034c32f88

・Pythonでもっと自由を　python-pptx
    　https://www.shibutan-bloomers.com/category/python/外部ライブラリ/python-pptx/

・PythonでPowerpoint(パワーポイント)を自動化【テキストの装飾】
    https://hogelog.com/python/python-powerpoint-2.html
    
・Quiita pythonでPowerPoint のフォント情報をcsvに出力
    https://qiita.com/kurokurokurotti/items/b9f0c7fe181bd0c8f85d
   
・似非プログラマの備忘録 python でPowerpointを自動化
    https://hogelog.com/python
 
------------------------------------------------------------------
"""
