"""
    Module: pySRIMfit
    File:   __init__.py
    Descr.: SRIMfit by python
    Update: 220403 build start Ay
            220515 v.1.0 done
    Last:   230310 v.1.1 refactored
   
@author: Ayoshida(RIKEN,RNC)
----.---.---.---.---.---.---.---.---.---.---.---.---.---.---.---.
Provides:
  * 
How To use:
  import pySRIMfit as pysr
  help(pysr.funcname)

Available subpackages
---------------------
  non

Utilities
---------
  non

__version__ = 1.1.0
----.---.---.---.---.---.---.---.---.---.---.---.---"""

#from pathlib import Path
#import sys
#sys.path.append(str(Path(__file__).resolve().parent.parent))
#print(sys.path)

from . srimWSmisc   import (_Err, atomSym,
                    srimWS_TrgCmp, srimWS_Uid, 
                    srimWSname_AZT, srimWSname_ASymT, srimWSname_BmT,
                    srimWSname_split, srimWSdump, 
                    srimWSplot_LET, srimWSplot_RngSt,
                    srimWSplot_LETcmp, srimWSplot_RngStcmp)
from . srimWSbase   import (srimWS)

from . srimWBmisc   import (srimWBdump)
from . srimWBbase   import (srimWB)

from . srimWScnv    import (srimWScnv)
from . srimWBcnv    import (srimWBcnv_xl2pick, srimWBcnv_xl2srWB, 
                            srimWBcnv_srWB2xl, srimWBcnv_pick2xl, 
                            srimWBcnv_out2xl,  srimWBcnv_csv2xl)
from . e5SRIMbase   import (e5SRIM)
from . mySRIMbase   import (mySRIM)
from . srimWBupd    import (srimWBupdater)


__all__ = [ 
# rawdmp_fio
    '_Err','atomSym',
    'srimWS_TrgCmp', 'srimWS_Uid',
    'srimWSname_AZT','srimWSname_ASymT','srimWSname_BmT',
    'srimWSname_split', 'srimWSdump',
    'srimWSplot_LET', 'srimWSplot_RngSt',
    'srimWSplot_LETcmp', 'srimWSplot_RngStcmp',

    'srimWS',

    'srimWBdump',

    'srimWB',

    'srimWScnv',

    'srimWBcnv_xl2pick', 'srimWBcnv_xl2srWB', 
    'srimWBcnv_srWB2xl', 'srimWBcnv_pick2xl', 
    'srimWBcnv_out2xl',  'srimWBcnv_csv2xl',

    'e5SRIM',

    'mySRIM',

    'srimWBupdater'
]


""" ========================================================
-----------------------
    __init__.py に関して
-----------------------
r**) Python__init__.py入門：基本から応用まで分かりやすく解説
    https://ya6mablog.com/how-to-use-init-py/
    
    
r01) Qiita 2020 Python の __init__.py とは何なのか
◎   https://qiita.com/msi/items/d91ea3900373ff8b09d7
   Python3.3 で新しく追加された「名前空間パッケージ (Namespace Packages)」のための機能で、
   「通常のパッケージ (Regular packages)」では、__init__.py が必要です。
   __init__.py は、ディレクトリ名をモジュール名(あるいは明示的な「名前空間」)として
   マッピングするためのファイル

r02) Qiita 2022 Pythonの相対パスimportを理解する
    https://qiita.com/u943425f/items/bd94a30b52c9296e942d

r03) Qiita 2016 Python 3でのファイルのimportのしかたまとめ
    https://qiita.com/karadaharu/items/37403e6e82ae4417d1b3
    ひとつ上のディレクトリにあるファイル
    from pathlib import Path
    import sys    
    sys.path.append(str(Path(__file__).resolve().parent.parent))
    import mother

r04) nlmk Pythonの相対インポートで上位ディレクトリ・サブディレクトリを指定
    https://note.nkmk.me/python-relative-import/

r**) Python__init__.py入門：基本から応用まで分かりやすく解説
    https://ya6mablog.com/how-to-use-init-py/
    
r**) Zenn Python でインポートがうまくいかないとき
    https://zenn.dev/suyaa/articles/cbc691a4797e61
    
r**) [Python]別ディレクトリの自作モジュールをimport
    https://fuji-pocketbook.net/another-dir-module/
    sitepackagesのなかに拡張子.pthのファイルを作成してパスを記入
    
    ========================================================"""
