# -*- coding: utf-8 -*-
"""
    Module: pySRIMfit:xlwings
    File:   xw_pySRIMfit.py
    Descr.: pySRIMfit wrapper for xlwings
        UDF names are identical with SRIMfit.xlam, although
        '=srFuncName()' --> '=pysrFuncName()' these changes needed.
    Update: 220417 build start Ay
    Last:   2204** v1.0 
@author: Ayoshida(RIKEN,RNC)
"""

""" import: standard lib.       
----.---.---.---.---.---.---.---.---.---.---.---.---"""
import  xlwings         as  xw
import  os
#import numpy            as np
#import scipy.optimize   as sopt

""" import: my custom lib.       
----.---.---.---.---.---.---.---.---.---.---.---.---"""
#import  aypylib             as ay 

""" import: from this package   
----.---.---.---.---.---.---.---.---.---.---.---.---"""
import  pySRIMfit       as  pysr

"""
    Global var. in this module
"""
srWB    = None      # srimWB instance
srWBpath= None      # MySRIMwb load path

_DBG  = False       # for debug msg. CTRL

""" .===.===.===.===.===.===.===.===.===.===.===.===.===.===.===.===
        pySRIMfit startup 
====.===.===.===.===.===.===.===.===.===.===.===.===.===.===.=== === """
def srWB_init():
    global srWB,srWBpath    # need this as this func changes global variable
    
    srWB = pysr.srimWB()
    # MySRIMwb should be located in a same folder of this script
    srWBpath = os.path.dirname(__file__)
    #print('curdir=',srWBpath)
    srWB.loadWS( srWBpath )
    print(srWB)
    print('srWB_init: Fin')

""" .===.===.===.===.===.===.===.===.===.===.===.===.===.===.===.===
        User Defind Funcs (UDF)
====.===.===.===.===.===.===.===.===.===.===.===.===.===.===.=== === """

""" misc. info.
----.---.---.---.---.---.---.---.---.---.---.---.---"""

@xw.func
def pysrMcrVer() -> str:
    """ Macro Version number """
    if(srWB is None): srWB_init()
    return( srWB.VerInfo )

@xw.func
def pysrMcrPath() -> str:
    """ Macro Path = MySRIMwb load path """
    if(srWB is None): srWB_init()
    return( srWBpath )

@xw.func
def pysrMcrWBname() -> str:
    """  MySRIMwb file name """
    if(srWB is None): srWB_init()
    return( srWB.WBname )

@xw.func
def pysrMcrWBdir() -> str:
    """  MySRIMwb file path """
    if(srWB is None): srWB_init()
    return( srWBpath )

@xw.func
def pysrMcrWScount() -> int:
    """  Num.of registerd Work Sheets  """
    if(srWB is None): srWB_init()
    return( srWB.nWS )

@xw.func
def pysrMcrWSlist() -> list:
    """  List of registerd Work Sheet names """
    if(srWB is None): srWB_init()
    # TODO 要動作確認
    return( srWB.WSname.join(',') )

@xw.func
def pysrMcr_WSlist() -> list:
    """  List of registerd Work Sheet names """
    if(srWB is None): srWB_init()
    # TODO 縦書きリストで返す
    pass

@xw.func
@xw.arg('Z', doc='Z-number')
def pysrElmNm(Z:int) -> str:
    """ Atomic Element name     """
    atms = pysr.atomSym()
    iZ= int(Z)
    sym = atms.sym(iZ)
    return( sym )

""" SRIM Work Sheet info.
----.---.---.---.---.---.---.---.---.---.---.---.---"""

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
def pysrInfoVer(WS:str) -> int:
    """ SRIM ver# of this WS    """
    if(srWB is None): srWB_init()
    return( srWB.IonZ(WS) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
def pysrInfoIonZ(WS:str) -> int:
    """ Beam Z-number   """
    if(srWB is None): srWB_init()
    return( srWB.IonZ(WS) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
def pysrInfoIonA(WS:str) -> int:
    """ Beam A-number   """
    if(srWB is None): srWB_init()
    return( srWB.IonA(WS) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
def pysrInfoTrgName(WS:str) -> str:
    """ Target name     """
    if(srWB is None): srWB_init()
    return( srWB.Target(WS) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
def pysrInfoTrgNameL(WS:str) -> str:
    """ Target name     """
    if(srWB is None): srWB_init()
    return( srWB.TargetNm(WS) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
def pysrInfoTrgDens(WS:str) -> float:
    """ Target Density in [g/cm3]     """
    if(srWB is None): srWB_init()
    return( srWB.TrgDens(WS) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
def pysrInfoTrgDensA(WS:str) -> float:
    """ Target Density in [atoms/cm3]     """
    if(srWB is None): srWB_init()
    return( srWB.TrgDensAtm(WS) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
def pysrInfoBrgC(WS:str) -> float:
    """ Bragg Correction in [%]     """
    if(srWB is None): srWB_init()
    return( srWB.BraggC(WS) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
def pysrInfoTrgPtbl(WS:str) -> float:
    """ If Gas, its Pressure [Pa] conciderd in this WS  """
    if(srWB is None): srWB_init()
    return( srWB.GasPtbl(WS) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
def pysrInfoTrgTtbl(WS:str) -> float:
    """ If Gas, its Temperature [degC] conciderd in this WS  """
    if(srWB is None): srWB_init()
    return( srWB.GasTtbl(WS) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
def pysrInfoIsGas(WS:str) -> bool:
    """ This WS is Gas target ?     """
    if(srWB is None): srWB_init()
    return( srWB.Gas(WS) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
@xw.arg('no', doc='Target composiotion list; line #')
def pysrInfoTgCmAtmNm(WS:str, no:int=0) -> str:
    """ Target composition table: Atomic Name """
    if(srWB is None): srWB_init()
    cmpL = srWB.TrgCmp(WS)
    cmp  = cmpL[no]
    return( cmp[ srWB._srimWS_TrgCmp[0] ] )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
@xw.arg('no', doc='Target composiotion list; line #')
def pysrInfoTgCmAtmNo(WS:str, no:int=0) -> int:
    """ Target composition table: Atomic Number """
    if(srWB is None): srWB_init()
    cmpL = srWB.TrgCmp(WS)
    cmp  = cmpL[no]
    return( cmp[ srWB._srimWS_TrgCmp[1] ] )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
@xw.arg('no', doc='Target composiotion list; line #')
def pysrInfoTgCmAtmPct(WS:str, no:int=0) -> float:
    """ Target composition table: Atomic [%] """
    if(srWB is None): srWB_init()
    cmpL = srWB.TrgCmp(WS)
    cmp  = cmpL[no]
    return( cmp[ srWB._srimWS_TrgCmp[2] ] )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
@xw.arg('no', doc='Target composiotion list; line #')
def pysrInfoTgCmMassPct(WS:str, no:int=0) -> float:
    """ Target composition table: Mass [%] """
    if(srWB is None): srWB_init()
    cmpL = srWB.TrgCmp(WS)
    cmp  = cmpL[no]
    return( cmp[ srWB._srimWS_TrgCmp[3] ] )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
def pysrInfoWScorded(WS:str) -> str:
    """ This WS is Corded by    """
    if(srWB is None): srWB_init()
    return( srWB.Corded(WS) )

@xw.func
@xw.arg('Uid',doc='LET unit ID#(0..8)')
def pysrLETUNm(Uid:int) -> str:
    """ LET Unit Name   """
    if(srWB is None): srWB_init()
    return( srWB.LETunit(Uid) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
def pysrLETCnvF(WS:str,Uid:int) -> str:
    """ LET unit Conversion Factor  """
    if(srWB is None): srWB_init()
    return( srWB.LETcnvF(WS,Uid) )

""" SRIM Work Sheet : basic functions
----.---.---.---.---.---.---.---.---.---.---.---.---"""

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
def pysrum2mg(WS:str,Tum:float) -> float:
    """ Trg. thick. conv. [micron] -> [mg/cm2]  """
    if(srWB is None): srWB_init()
    return( srWB.um2mg(WS,Tum) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
def pysrmg2um(WS:str,Tmg:float) -> float:
    """ Trg. thick. conv. [mg/cm2] -> [micron]  """
    if(srWB is None): srWB_init()
    return( srWB.mg2um(WS,Tmg) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
def pysrMinE(WS:str) -> float:
    """ min E [MeV/A] in this WS    """
    if(srWB is None): srWB_init()
    return( srWB.minE(WS) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
def pysrMaxE(WS:str) -> float:
    """ max E [MeV/A] in this WS    """
    if(srWB is None): srWB_init()
    return( srWB.maxE(WS) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
@xw.arg('Uid',doc='LET unit ID#(0..8)')
def pysrMinLETe(WS:str, Uid:int) -> float:
    """ min LET_elec. [MeV/(mg/cm2)] in this WS    """
    if(srWB is None): srWB_init()
    return( srWB.minLETe(WS,Uid) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
def pysrMaxLETe(WS:str, Uid:int) -> float:
    """ max LET_elec. [MeV/(mg/cm2)] in this WS    """
    if(srWB is None): srWB_init()
    return( srWB.maxLETe(WS,Uid) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
def pysrMinLETn(WS:str, Uid:int) -> float:
    """ min LET_nucl. [MeV/(mg/cm2)] in this WS    """
    if(srWB is None): srWB_init()
    return( srWB.minLETn(WS,Uid) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
def pysrMaxLETn(WS:str, Uid:int) -> float:
    """ max LET_nucl. [MeV/(mg/cm2)] in this WS    """
    if(srWB is None): srWB_init()
    return( srWB.maxLETn(WS,Uid) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
def pysrMinLETt(WS:str, Uid:int) -> float:
    """ min LET_total [MeV/(mg/cm2)] in this WS    """
    if(srWB is None): srWB_init()
    return( srWB.minLETt(WS,Uid) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
def pysrMaxLETt(WS:str, Uid:int) -> float:
    """ max LET_total [MeV/(mg/cm2)] in this WS    """
    if(srWB is None): srWB_init()
    return( srWB.maxLETt(WS,Uid) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
def pysrMinRng(WS:str) -> float:
    """ min Range [um] in this WS    """
    if(srWB is None): srWB_init()
    return( srWB.minRng(WS) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
def pysrMaxRng(WS:str) -> float:
    """ max Range [um] in this WS    """
    if(srWB is None): srWB_init()
    return( srWB.maxRng(WS) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
def pysrMinStLng(WS:str) -> float:
    """ min Straggling Langitudinal [um] in this WS    """
    if(srWB is None): srWB_init()
    return( srWB.minStLng(WS) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
def pysrMaxStLng(WS:str) -> float:
    """ max Straggling Langitudinal [um] in this WS    """
    if(srWB is None): srWB_init()
    return( srWB.maxStLng(WS) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
def pysrMinStLtr(WS:str) -> float:
    """ min Straggling Lateral [um] in this WS    """
    if(srWB is None): srWB_init()
    return( srWB.minStLtr(WS) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
def pysrMaxStLtr(WS:str) -> float:
    """ max Straggling Lateral [um] in this WS    """
    if(srWB is None): srWB_init()
    return( srWB.maxStLtr(WS) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
@xw.arg('E',  doc='Energy [MeV/A]')
@xw.arg('Uid',doc='LET unit ID#(0..8)')
def pysrE2LETe(WS:str,E:float,Uid:int) -> float:
    """ E -> LET_elec [Uid unit] """
    if(srWB is None): srWB_init()
    return( srWB.LETe(WS,E,Uid) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
@xw.arg('E',  doc='Energy [MeV/A]')
@xw.arg('Uid',doc='LET unit ID#(0..8)')
def pysrE2LETn(WS:str,E:float,Uid:int) -> float:
    """ E -> LET_nucl [Uid unit] """
    if(srWB is None): srWB_init()
    return( srWB.LETn(WS,E,Uid) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
@xw.arg('E',  doc='Energy [MeV/A]')
@xw.arg('Uid',doc='LET unit ID#(0..8)')
def pysrE2LETt(WS:str,E:float,Uid:int) -> float:
    """ E -> LET_total [Uid unit] """
    if(srWB is None): srWB_init()
    return( srWB.LETt(WS,E,Uid) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
@xw.arg('Le', doc='LET_elec [Uid unit]')
@xw.arg('Uid',doc='LET unit ID#(0..8)')
@xw.arg('Ehl',doc='{+1|-1} search at E-{higher|lower} side of Bragg peak')
def pysrLETe2E(WS:str,Le:float,Uid:int,Ehl:int) -> float:
    """ LET_elec [Uid unit] -> E [MeV/A] """
    if(srWB is None): srWB_init()
    Hside = False if(Ehl == -1) else True
    return( srWB.LETe2E(WS,Le,Uid=Uid,Hside=Hside) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
@xw.arg('Ln', doc='LET_elec [Uid unit]')
@xw.arg('Uid',doc='LET unit ID#(0..8)')
@xw.arg('Ehl',doc='{+1|-1} search at E-{higher|lower} side of Bragg peak')
def pysrLETn2E(WS:str,Ln:float,Uid:int,Ehl:int) -> float:
    """ LET_nucl [Uid unit] -> E [MeV/A] """
    if(srWB is None): srWB_init()
    Hside = False if(Ehl == -1) else True
    return( srWB.LETn2E(WS,Ln,Uid=Uid,Hside=Hside) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
@xw.arg('Lt', doc='LET_total [Uid unit]')
@xw.arg('Uid',doc='LET unit ID#(0..8)')
@xw.arg('Ehl',doc='{+1|-1} search at E-{higher|lower} side of Bragg peak')
def pysrLETt2E(WS:str,Lt:float,Uid:int,Ehl:int) -> float:
    """ LET_elec [Uid unit] -> E [MeV/A] """
    if(srWB is None): srWB_init()
    Hside = False if(Ehl == -1) else True
    return( srWB.LETt2E(WS,Lt,Uid=Uid,Hside=Hside) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
def pysrMaxLETe2E(WS:str) -> float:
    """ max LET_elec -> E [MeV/A] """
    if(srWB is None): srWB_init()
    return( srWB.maxLETe2E(WS) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
def pysrMaxLETn2E(WS:str) -> float:
    """ max LET_nucl -> E [MeV/A] """
    if(srWB is None): srWB_init()
    return( srWB.maxLETn2E(WS) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
def pysrMaxLETt2E(WS:str) -> float:
    """ max LET_total -> E [MeV/A] """
    if(srWB is None): srWB_init()
    return( srWB.maxLETt2E(WS) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
@xw.arg('E',  doc='Energy [MeV/A]')
def pysrE2Rng(WS:str,E:float) -> float:
    """ E [MeV/A] -> Range [um] """
    if(srWB is None): srWB_init()
    return( srWB.Rng(WS,E) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
@xw.arg('Rng',doc='Range [um]')
def pysrRng2E(WS:str,Rng:float) -> float:
    """ Range [um] -> E [MeV/A]     """
    if(srWB is None): srWB_init()
    return( srWB.Rng2E(WS,Rng) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
@xw.arg('E',  doc='Energy [MeV/A]')
def pysrE2StLng(WS:str,E:float) -> float:
    """ E [MeV/A] -> Straggling Longitudinal [um] """
    if(srWB is None): srWB_init()
    return( srWB.StLng(WS,E) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
@xw.arg('St', doc='Straggling Longitudinal [um]')
def pysrStLng2E(WS:str,St:float) -> float:
    """ Straggling Longitudinal [um] -> E [MeV/A]     """
    if(srWB is None): srWB_init()
    return( srWB.StLng2E(WS,St) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
@xw.arg('E',  doc='Energy [MeV/A]')
def pysrE2StLtr(WS:str,E:float) -> float:
    """ E [MeV/A] -> Straggling Lateral [um] """
    if(srWB is None): srWB_init()
    return( srWB.StLtr(WS,E) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
@xw.arg('St', doc='Straggling Lateral [um]')
def pysrStLtr2E(WS:str,St:float) -> float:
    """ Straggling Lateral [um] -> E [MeV/A]     """
    if(srWB is None): srWB_init()
    return( srWB.StLtr2E(WS,St) )

""" SRIM Work Sheet : application functions
----.---.---.---.---.---.---.---.---.---.---.---.---"""

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
@xw.arg('Pa', doc='Gass Pressure [Pa]')
@xw.arg('dgC',doc='Gass Temperature [degC]')
def pysrThkStd(WS:str,Pa:float,dgC:float) -> float:
    """ Gas thick. conv. factor to "Standard P,T" of this SRIM table """
    if(srWB is None): srWB_init()
    return( srWB.ThkStdGas(WS,Pa,dgC) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
@xw.arg('E0', doc='Beam Energy [MeV/A]')
@xw.arg('Tum',doc='Target thick. [micron]')
def pysrEnew(WS:str,E0:float,Tum:float) -> float:
    """ E0 -> Thick[um] -> E1:new-energy """
    if(srWB is None): srWB_init()
    return( srWB.Enew(WS,E0,Tum) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
@xw.arg('E1', doc='Beam Energy [MeV/A]')
@xw.arg('Tum',doc='Target thick. [micron]')
def pysrEold(WS:str,E1:float,Tum:float) -> float:
    """ E0:old-energy <- Thick[um] <- E1 """
    if(srWB is None): srWB_init()
    return( srWB.Eold(WS,E1,Tum) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
@xw.arg('E0', doc='Beam Energy [MeV/A]')
@xw.arg('Tmm',doc='Target thick. [mm]')
@xw.arg('Pa', doc='Gass Pressure [Pa]')
@xw.arg('dgC',doc='Gass Temperature [degC]')
def pysrEnewGas(WS:str,E0:float,Tmm:float, Pa:float, dgC:float) -> float:
    """ E0 -> Thick[um],Pa,degC -> E1:new-energy """
    if(srWB is None): srWB_init()
    return( srWB.EnewGas(WS,E0,Tmm,Pa,dgC) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
@xw.arg('E1', doc='Beam Energy [MeV/A]')
@xw.arg('Tmm',doc='Target thick. [mm]')
@xw.arg('Pa', doc='Gass Pressure [Pa]')
@xw.arg('dgC',doc='Gass Temperature [degC]')
def pysrEoldGas(WS:str,E1:float,Tmm:float, Pa:float, dgC:float) -> float:
    """ E0:old-energy <- Thick[um],Pa,degC <- E1 """
    if(srWB is None): srWB_init()
    return( srWB.EoldGas(WS,E1,Tmm,Pa,dgC) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
@xw.arg('Pa', doc='Gass Pressure [Pa]')
@xw.arg('dgC',doc='Gass Temperature [degC]')
def pysrThkStd(WS:str,Pa:float, dgC:float) -> float:
    """ Gas thick. conv. factor to "Standard P,T" of this SRIM table """
    if(srWB is None): srWB_init()
    return( srWB.ThkStd(WS,Pa,dgC) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
@xw.arg('Pa', doc='Gass Pressure [Pa]')
@xw.arg('dgC',doc='Gass Temperature [degC]')
@xw.arg('Tmm',doc='Target thick. [mm]')
def pysrmm2umGas(WS:str,Tmm:float, Pa:float, dgC:float) -> float:
    """ Gas Thickness convertion : mm -> micron """
    if(srWB is None): srWB_init()
    return( srWB.mm2umGas(WS,Tmm,Pa,dgC) )

@xw.func
@xw.arg('WS', doc='SRIM work-sheet name')
@xw.arg('Pa', doc='Gass Pressure [Pa]')
@xw.arg('dgC',doc='Gass Temperature [degC]')
@xw.arg('Tum',doc='Target thick. [micron]')
def pysrum2mmGas(WS:str,Tum:float, Pa:float, dgC:float) -> float:
    """ Gas Thickness convertion : micron -> mm """
    if(srWB is None): srWB_init()
    return( srWB.um2mmGas(WS,Tum,Pa,dgC) )

""" SRIM Work Sheet : Two SrimWS combination func.
----.---.---.---.---.---.---.---.---.---.---.---.---"""

@xw.func
@xw.arg('WS1', doc='SRIM work-sheet name')
@xw.arg('WS2', doc='SRIM work-sheet name')
@xw.arg('E10', doc='Beam Energy [MeV/A] for WS1')
def pysrRng_eq_Eold(WS1:str,WS2:str, E10:float) -> float:
    """ Range equivalent Eold
        ie.) get E20 of 
                 Rng(WS1,E10) == Rng(WS2,E20) """
    if(srWB is None): srWB_init()
    return( srWB.Rng_eq_Eold(WS1,WS2,E10) )

@xw.func
@xw.arg('WS1', doc='SRIM work-sheet name')
@xw.arg('WS2', doc='SRIM work-sheet name')
@xw.arg('E10', doc='Beam Energy [MeV/A] for WS1')
@xw.arg('Th1', doc='Target thick. [micron] for WS1')
def pysrEnew_eq_Th(WS1:str,WS2:str, E10:float, Th1:float) -> float:
    """ Enew equivalent Thickness
        ie.) get Th2 of
                 Enew(WS1,E10,Th1) 
              == Enew(WS2,E20 == E10,Th2) """
    if(srWB is None): srWB_init()
    return( srWB.Enew_eq_Th(WS1,WS2,E10,Th1) )

@xw.func
@xw.arg('WS1', doc='SRIM work-sheet name')
@xw.arg('WS2', doc='SRIM work-sheet name')
@xw.arg('E10', doc='Beam Energy [MeV/A] for WS1')
@xw.arg('Th1', doc='Target thick. [micron] for WS1')
def pysrEnewt_eq_Th(WS1:str,WS2:str, E10:float, Th1:float) -> float:
    """ Enew * A:totalE equivalent Thickness
        ie.) get Th2 of 
               Enew(WS1,E10,Th1) * A1 
            == Enew(WS2,E20 == E10,Th2) * A2 """
    if(srWB is None): srWB_init()
    return( srWB.Enewt_eq_Th(WS1,WS2,E10,Th1) )

@xw.func
@xw.arg('WS1', doc='SRIM work-sheet name')
@xw.arg('WS2', doc='SRIM work-sheet name')
@xw.arg('E10', doc='Beam Energy [MeV/A] for WS1')
@xw.arg('Th1', doc='Target thick. [micron] for WS1')
def pysrEnew_eq_Eold(WS1:str,WS2:str, E10:float, Th1:float) -> float:
    """ Enew equivalent Eold
        ie.) get E20 of
                 Enew(WS1,E10,Th1)
              == Enew(WS2,E20,Th2 == Th1) """
    if(srWB is None): srWB_init()
    return( srWB.Enew_eq_Eold(WS1,WS2,E10,Th1) )

@xw.func
@xw.arg('WS1', doc='SRIM work-sheet name')
@xw.arg('WS2', doc='SRIM work-sheet name')
@xw.arg('E10', doc='Beam Energy [MeV/A] for WS1')
@xw.arg('Th1', doc='Target thick. [micron] for WS1')
def pysrEnewt_eq_Eold(WS1:str,WS2:str, E10:float, Th1:float) -> float:
    """ E Enew * A:totalE equivalent Eold
        ie.) get E20 of 
                 Enew(WS1,E10,Th1) * A1
             ==  Enew(WS2,E20,Th2 == Th1) * A2 """
    if(srWB is None): srWB_init()
    return( srWB.Enewt_eq_Eold(WS1,WS2,E10,Th1) )

@xw.func
@xw.arg('WS1', doc='SRIM work-sheet name')
@xw.arg('WS2', doc='SRIM work-sheet name')
@xw.arg('E10', doc='Beam Energy [MeV/A] for WS1')
@xw.arg('Th1', doc='Target thick. [micron] for WS1')
def pysrDE_eq_Th(WS1:str,WS2:str, E10:float, Th1:float) -> float:
    """ dE [MeV/A] equivalent Thickness
        ie.) get Th2 of 
            dE1 = (E10        - Enew(WS1,E10,Th1)) 
         == dE2 = (E20 == E10 - Enew(WS2,E20,Th2)) """
    if(srWB is None): srWB_init()
    return( srWB.dE_eq_Th(WS1,WS2,E10,Th1) )

@xw.func
@xw.arg('WS1', doc='SRIM work-sheet name')
@xw.arg('WS2', doc='SRIM work-sheet name')
@xw.arg('E10', doc='Beam Energy [MeV/A] for WS1')
@xw.arg('Th1', doc='Target thick. [micron] for WS1')
def pysrDEt_eq_Th(WS1:str,WS2:str, E10:float, Th1:float) -> float:
    """ dE * A:totalE equivalent Thickness
        ie.) get Th2 of 
            dE1 = (E10        - Enew(WS1,E10,Th1)) * A1 
         == dE2 = (E20 == E10 - Enew(WS2,E20,Th2)) * A2 """
    if(srWB is None): srWB_init()
    return( srWB.dEt_eq_Th(WS1,WS2,E10,Th1) )

@xw.func
@xw.arg('WS1', doc='SRIM work-sheet name')
@xw.arg('WS2', doc='SRIM work-sheet name')
@xw.arg('E10', doc='Beam Energy [MeV/A] for WS1')
@xw.arg('Th1', doc='Target thick. [micron] for WS1')
@xw.arg('Ehl',doc='{+1|-1} search at E-{higher|lower} side of Bragg peak')
def pysrDEu_eq_Eold(WS1:str,WS2:str, E10:float, Th1:float, Ehl:int) -> float:
    """ Enew * A:totalE equivalent Eold
        ie.) get E20 of 
            dE1 = (E10 - Enew(WS1,E10,Th1)) 
         == dE2 = (E20 - Enew(WS2,E20,Th2 == Th1)) """
    if(srWB is None): srWB_init()
    Hside = False if(Ehl == -1) else True    
    return( srWB.dE_eq_Eold(WS1,WS2,E10,Th1,Hside) )

@xw.func
@xw.arg('WS1', doc='SRIM work-sheet name')
@xw.arg('WS2', doc='SRIM work-sheet name')
@xw.arg('E10', doc='Beam Energy [MeV/A] for WS1')
@xw.arg('Th1', doc='Target thick. [micron] for WS1')
@xw.arg('Ehl',doc='{+1|-1} search at E-{higher|lower} side of Bragg peak')
def pysrDEt_eq_Eold(WS1:str,WS2:str, E10:float, Th1:float, Ehl:int) -> float:
    """ Enew * A:totalE equivalent Eold
        ie.) get E20 of 
            dE1t = (E10 - Enew(WS1,E10,Th1)) * A1 
         == dE2t = (E20 - Enew(WS2,E20,Th2 == Th1)) * A2 """
    if(srWB is None): srWB_init()
    Hside = False if(Ehl == -1) else True    
    return( srWB.dEt_eq_Eold(WS1,WS2,E10,Th1,Hside) )

""" SRIM Work Sheet : Two SrimWS combination func. : Gas
----.---.---.---.---.---.---.---.---.---.---.---.---"""

@xw.func
@xw.arg('WS1',  doc='SRIM work-sheet name')
@xw.arg('WS2',  doc='SRIM work-sheet name')
@xw.arg('E10',  doc='Beam Energy [MeV/A] for WS1')
@xw.arg('Tmm1', doc='Target thick. [mm] for WS1')
@xw.arg('Pa1',  doc='Gass Pressure [Pa] for WS1')
@xw.arg('degC1',doc='Gass Temperature [degC] for WS1')
@xw.arg('Pa2',  doc='Gass Pressure [Pa] for WS2')
@xw.arg('degC2',doc='Gass Temperature [degC] for WS2')
def pysrEnew_eq_ThGas(WS1:str,WS2:str, 
                E10:float, Tmm1:float, Pa1:float, degC1:float,
                Pa2:float, degC2:float) -> float:
    """ Enew equivalent Thickness : Gas
        ie.) get Tmm2 of
                 EnewGas(WS1,E10,Tmm1,Pa1,degC1) 
              == EnewGas(WS2,E20 == E10,Tmm2,Pa2,degC2) """
    if(srWB is None): srWB_init()
    return( srWB.Enew_eq_ThGas(WS1,WS2,E10,Tmm1,Pa1,degC1,Pa2,degC2) )

@xw.func
@xw.arg('WS1',  doc='SRIM work-sheet name')
@xw.arg('WS2',  doc='SRIM work-sheet name')
@xw.arg('E10',  doc='Beam Energy [MeV/A] for WS1')
@xw.arg('Tmm1', doc='Target thick. [mm] for WS1')
@xw.arg('Pa1',  doc='Gass Pressure [Pa] for WS1')
@xw.arg('degC1',doc='Gass Temperature [degC] for WS1')
@xw.arg('Pa2',  doc='Gass Pressure [Pa] for WS2')
@xw.arg('degC2',doc='Gass Temperature [degC] for WS2')
def pysrEnewt_eq_ThGas(WS1:str,WS2:str, 
                E10:float, Tmm1:float, Pa1:float, degC1:float,
                Pa2:float, degC2:float) -> float:
    """ Enew * A:totalE equivalent Thickness : Gas
        ie.) get Tmm2 of 
                 EnewGas(WS1,E10,Tmm1,Pa1,degC1) * A1 
              == EnewGas(WS2,E20 == E10,Tmm2,Pa2,degC2) * A2 """
    if(srWB is None): srWB_init()
    return( srWB.Enewt_eq_ThGas(WS1,WS2,E10,Tmm1,Pa1,degC1,Pa2,degC2) )

@xw.func
@xw.arg('WS1',  doc='SRIM work-sheet name')
@xw.arg('WS2',  doc='SRIM work-sheet name')
@xw.arg('E10',  doc='Beam Energy [MeV/A] for WS1')
@xw.arg('Tmm1', doc='Target thick. [mm] for WS1')
@xw.arg('Pa1',  doc='Gass Pressure [Pa] for WS1')
@xw.arg('degC1',doc='Gass Temperature [degC] for WS1')
@xw.arg('Pa2',  doc='Gass Pressure [Pa] for WS2')
@xw.arg('degC2',doc='Gass Temperature [degC] for WS2')
def pysrEnew_eq_EoldGas(WS1:str,WS2:str, 
                E10:float, Tmm1:float, Pa1:float, degC1:float,
                Pa2:float, degC2:float) -> float:
    """ Enew equivalent Eold : Gas
        ie.) get E20 of
                 EnewGas(WS1,E10,Tmm1,Pa1,degC1)
              == EnewGas(WS2,E20,Tmm2 == Tmm1,Pa2,degC2) """
    if(srWB is None): srWB_init()
    return( srWB.Enew_eq_EoldGas(WS1,WS2,E10,Tmm1,Pa1,degC1,Pa2,degC2) )

@xw.func
@xw.arg('WS1',  doc='SRIM work-sheet name')
@xw.arg('WS2',  doc='SRIM work-sheet name')
@xw.arg('E10',  doc='Beam Energy [MeV/A] for WS1')
@xw.arg('Tmm1', doc='Target thick. [mm] for WS1')
@xw.arg('Pa1',  doc='Gass Pressure [Pa] for WS1')
@xw.arg('degC1',doc='Gass Temperature [degC] for WS1')
@xw.arg('Pa2',  doc='Gass Pressure [Pa] for WS2')
@xw.arg('degC2',doc='Gass Temperature [degC] for WS2')
def pysrEnewt_eq_EoldGas(WS1:str,WS2:str, 
                E10:float, Tmm1:float, Pa1:float, degC1:float,
                Pa2:float, degC2:float) -> float:
    """ Enew * A:totalE equivalent Eold : Gas
        ie.) get E20 of
                 EnewGas(WS1,E10,Tmm1,Pa1,degC1) * A1
              == EnewGas(WS2,E20,Tmm2 == Tmm1,Pa2,degC2) *A2 """
    if(srWB is None): srWB_init()
    return( srWB.Enewt_eq_EoldGas(WS1,WS2,E10,Tmm1,Pa1,degC1,Pa2,degC2) )

@xw.func
@xw.arg('WS1',  doc='SRIM work-sheet name')
@xw.arg('WS2',  doc='SRIM work-sheet name')
@xw.arg('E10',  doc='Beam Energy [MeV/A] for WS1')
@xw.arg('Tmm1', doc='Target thick. [mm] for WS1')
@xw.arg('Pa1',  doc='Gass Pressure [Pa] for WS1')
@xw.arg('degC1',doc='Gass Temperature [degC] for WS1')
@xw.arg('Pa2',  doc='Gass Pressure [Pa] for WS2')
@xw.arg('degC2',doc='Gass Temperature [degC] for WS2')
def pysrDE_eq_ThGas(WS1:str,WS2:str, 
                E10:float, Tmm1:float, Pa1:float, degC1:float,
                Pa2:float, degC2:float) -> float:
    """ dE [MeV/A] equivalent Thickness : Gas
        ie.) get Tmm2 of 
            dE1 = (E10        - EnewGas(WS1,E10,Tmm1,Pa1,degC1)) 
         == dE2 = (E20 == E10 - EnewGas(WS2,E20,Tmm2,Pa2,degC2))) """
    if(srWB is None): srWB_init()
    return( srWB.dE_eq_ThGas(WS1,WS2,E10,Tmm1,Pa1,degC1,Pa2,degC2) )

@xw.func
@xw.arg('WS1',  doc='SRIM work-sheet name')
@xw.arg('WS2',  doc='SRIM work-sheet name')
@xw.arg('E10',  doc='Beam Energy [MeV/A] for WS1')
@xw.arg('Tmm1', doc='Target thick. [mm] for WS1')
@xw.arg('Pa1',  doc='Gass Pressure [Pa] for WS1')
@xw.arg('degC1',doc='Gass Temperature [degC] for WS1')
@xw.arg('Pa2',  doc='Gass Pressure [Pa] for WS2')
@xw.arg('degC2',doc='Gass Temperature [degC] for WS2')
def pysrDEt_eq_ThGas(WS1:str,WS2:str, 
                E10:float, Tmm1:float, Pa1:float, degC1:float,
                Pa2:float, degC2:float) -> float:
    """ dE * A:totalE equivalent Thickness : Gas
        ie.) get Tmm2 of 
            dE1 = (E10        - EnewGas(WS1,E10,Tmm1)) * A1 
         == dE2 = (E20 == E10 - EnewGas(WS2,E20,Tmm2)) * A2 """
    if(srWB is None): srWB_init()
    return( srWB.dEt_eq_ThGas(WS1,WS2,E10,Tmm1,Pa1,degC1,Pa2,degC2) )

@xw.func
@xw.arg('WS1',  doc='SRIM work-sheet name')
@xw.arg('WS2',  doc='SRIM work-sheet name')
@xw.arg('E10',  doc='Beam Energy [MeV/A] for WS1')
@xw.arg('Tmm1', doc='Target thick. [mm] for WS1')
@xw.arg('Pa1',  doc='Gass Pressure [Pa] for WS1')
@xw.arg('degC1',doc='Gass Temperature [degC] for WS1')
@xw.arg('Pa2',  doc='Gass Pressure [Pa] for WS2')
@xw.arg('degC2',doc='Gass Temperature [degC] for WS2')
def pysrDE_eq_EoldGas(WS1:str,WS2:str, 
                E10:float, Tmm1:float, Pa1:float, degC1:float,
                Pa2:float, degC2:float) -> float:
    """ Enew * A:totalE equivalent Eold : Gas
        ie.) get E20 of 
            dE1 = (E10 - EnewGas(WS1,E10,Tmm1,Pa1,degC1)) 
         == dE2 = (E20 - EnewGas(WS2,E20,Tmm2 == Tmm1,Pa2,degC2)) """
    if(srWB is None): srWB_init()
    return( srWB.dE_eq_EoldGas(WS1,WS2,E10,Tmm1,Pa1,degC1,Pa2,degC2) )

@xw.func
@xw.arg('WS1',  doc='SRIM work-sheet name')
@xw.arg('WS2',  doc='SRIM work-sheet name')
@xw.arg('E10',  doc='Beam Energy [MeV/A] for WS1')
@xw.arg('Tmm1', doc='Target thick. [mm] for WS1')
@xw.arg('Pa1',  doc='Gass Pressure [Pa] for WS1')
@xw.arg('degC1',doc='Gass Temperature [degC] for WS1')
@xw.arg('Pa2',  doc='Gass Pressure [Pa] for WS2')
@xw.arg('degC2',doc='Gass Temperature [degC] for WS2')
def pysrDEt_eq_EoldGas(WS1:str,WS2:str, 
                E10:float, Tmm1:float, Pa1:float, degC1:float,
                Pa2:float, degC2:float) -> float:
    """ Enew * A:totalE equivalent Eold : Gas
        ie.) get E20 of 
            dE1 = (E10 - EnewGas(WS1,E10,Tmm1,Pa1,degC1)) * A1
         == dE2 = (E20 - EnewGas(WS2,E20,Tmm2 == Tmm1,Pa2,degC2)) * A2 """
    if(srWB is None): srWB_init()
    return( srWB.dEt_eq_EoldGas(WS1,WS2,E10,Tmm1,Pa1,degC1,Pa2,degC2) )

""" ========================================================
            Test Main
    ========================================================"""
if __name__ == '__main__':

    SwTest1 = False     # test xwfitGa() auto

    """ Test1: xwfitGa() auto mode
        ----------------------------------------------      """
    if(SwTest1):
        pass

    print('@@main: -- Fin ---')

    
""" References: ============================================
// xlwings //

r#101) Qiita2021 xlwingsトラブルシューティング
    https://qiita.com/k_maki/items/099699e4313a5d61da01





* xlwings: VBA: ユーザー定義関数 (UDFs)
  https://docs.xlwings.org/ja/latest/udfs.html
* Qiita 2019 ExcelにPythonが搭載？その後 - xlwings を使おう
　 https://qiita.com/yniji/items/b38bc312e860027108ac
* HatenaBlog 2019 VBAからPythonのコードを実行してみる(1,2)
　 https://celaeno42.hatenablog.com/entry/2018/12/03/191854
　 https://celaeno42.hatenablog.com/entry/2018/12/07/180351
* Newton Excel Bach, not (just) an Excel Blog
  xlwSciPy update for Python 3; April 9, 2018 
  https://newtonexcelbach.com/2018/04/09/xlwscipy-update-for-python-3/


 ============================================================ """
