# -*- coding: utf-8 -*-
"""
    Module: pySRIMfit
    File:   srimWB.py
    Descr.: def. class srimWB ; Work Book
    Update: 220403 build start Ay
            220515 v.1.0 done
    Last:   230424 v.1.1 refactored

    Future: ****** v.2.0 param = scalar | list
    
@author: Ayoshida(RIKEN,Nishina center)
"""

""" import: standard lib.       
----.---.---.---.---.---.---.---.---.---.---.---.---"""
import  pandas  as  pd
import  pickle 
import  os
import  operator    # used at .getWSnm_SpL()
import  copy        # used at ._interp_E_dE(())
import  numpy   as  np

""" import: my custom lib.       
----.---.---.---.---.---.---.---.---.---.---.---.---"""
#import  aypylib             as ay 

""" import: from this package   
----.---.---.---.---.---.---.---.---.---.---.---.---"""
from    pySRIMfit   import  (_Err, srimWS_Uid, 
                    srimWSname_AZT, srimWSname_ASymT, 
                    srimWSname_BmT, srimWSname_split,
                    srimWSplot_LETcmp, srimWSplot_RngStcmp, srimWBdump )

""" .===.===.===.===.===.===.===.===.===.===.===.===.===.===.===.===
        class srimWB
====.===.===.===.===.===.===.===.===.===.===.===.===.===.===.=== === """
class srimWB( object ):
    """ SRIM Work-Book (srWB) container & functions
Params. for constructor:
    non
Attributes:
  < book-info >
    .WBname     str     srWB binary file name (defa. MySRIMwb.pickle)
    .WBpath     str     srWB load path
    .nWS        int     num.of srWS registerd in this srWB
    .WSL        [srWS]  list of srim Work-Sheets registerd
    .WSnameL    [str]   list of srWS.wsName registered
    .WSnameLDB  [dict{'Bm':,'Tg':,..}]  dict list of srWS info
  < currently aim at sheet >
    .curWS      srWS    current srWS to get info.
    .curWSname  str     .curWS.wsName
Methods:
  < load / save / getter >
    .loadWS(path,fnm)       load MySRIMwb.pickle
    .saveWS(path,fnm)       save MySRIMwb.pickle
    .getWS(WSnm)            get srWS by wsName
    .getWSL(WSnmL)          get srWS list by wsName list
    .getWSnm_AZT(A,Z,Tg)        get srWS by ion A,Z & Targ.
    .getWSnm_ASymT(A,Sym,Tg)    get srWS by ion A,Z-symbol & Targ.
    .getWSnm_BmT(Bm,Tg)         get srWS by Beam & Targ.
    .getWSnmL(Bm,Tg)            get srWS name list [str,..]     
    .getWSnmLDB(Bm,Tg)          get srWS name list [{'Bm':*, 'Tg':*,..}]     
  < get srWS attributes by WSnm:work-sheet name >
    .wsName(WSnm)       srimWS name
    .SRIMver(WSnm)      SRIM version
    .IonZ(WSnm)         Beam Z-number
    .IonA(WSnm)         Beam A-number
    .Beam(WSnm)         Beam name eg) '84Kr'
    .Target(WSnm)       Target material (short name)
    .TargetNm(WSnm)     Target material (long name)
    .TrgDens(WSnm)      Target density [g/cm3]
    .TrgDensAtm(WSnm)   Target density [atoms/cm3]
    .TrgCmp(WSnm)       Target material composition table
    .BraggC(WSnm)       Bragg corection [%]
    .Gas(WSnm)          Target is Gas ?
    .GasPtbl(WSnm)      Gas Press.[Pa]  asuumed in this SRIM table.
    .GasTtbl(WSnm)      Gas Temp.[degC] asuumed.
    .Corded(WSnm)       Corded by info.
  < unit conversion >
    .ug2mg(WSnm,Tum)        micron -> mg
    .mg2um(WSnm,Tmg)        mg -> micron
    .mm2umGas(WSnm,Tmm,Pa,degC) gas thick. conv. mm -> micron
    .um2mmGas(WSnm,Tum,Pa,degC) gas thick. conv. micron -> mm
    .ThkStdGas(WSnm,Pa,degC)    gas thick. conv. factor for this srWS-values
    .LETunit(Uid)               LET unit name
    .LETcnvF(WSnm,Uid)          LET unit conversion factor used for;
                    LET(E,Uid) = LET(E;Uid=0) * CnvF(Uid)
  < get srWS attrib. : min|max value of the internal tables >
    .{min|max}E(WSnm)           min|max of E table
    .{min|max}LETe(WSnm,Uid)    min|max of LET_elec table in Uid unit
    .{min|max}LETn(WSnm,Uid)    min|max of LET_nucl table in Uid unit
    .{min|max}LETt(WSnm,Uid)    min|max of LET_tot  table in Uid unit
    .{min|max}Rng(WSnm)         min|max of Range table
    .{min|max}StLng(WSnm)       min|max of Straggling Longitudinal table
    .{min|max}StLtr(WSnm)       min|max of Straggling Lateral table
    .maxLETe2E(WSnm)            E at max LET_elec
    .maxLETn2E(WSnm)            E at max LET_nucl
    .maxLETt2E(WSnm)            E at max LET_total
  < functions(1): E --> Value >
    .LETe(WSnm,E,Uid) == .E2LETe(WSnm,E,Uid)  dE/dX_elec.
    .LETn(WSnm,E,Uid) == .E2LETn(WSnm,E,Uid)  dE/dX_nucl.
    .LETt(WSnm,E,Uid) == .E2LETt(WSnm,E,Uid)  dE/dX_tot = elec+nucl
    .Rng(WSnm,E)      == .E2Rng(WSnm,E)       Range
    .StLng(WSnm,E)    == .E2StLng(WSnm,E)     Straggling Longitudinal
    .StLtr(WSnm,E)    == .E2StLtr(WSnm,E)     Straggling Lateral
  < functions(2): Value --> E ; inverse functions >
    .LETe2E(WSnm,LETe,Uid,Hside) inv.func. of .LETe(E)
    .LETn2E(WSnm,LETn,Uid,Hside) inv.func. of .LETn(E)
    .LETt2E(WSnm,LETt,Uid,Hside) inv.func. of .LETt(E)
    .Rng2E(WSnm,Rng)             inv.func. of .Rng(E)
    .StLng2E(WSnm,StLng)         inv.func. of .StLng(E)
    .StLtr2E(WSnm,StLtr)         inv.func. of .StLtr(E)
  < functions(3): Enew / Eold / related funcs. >
    .Enew(   WSnm,E,Tum)            new energy through this target; thick=Tum
    .EnewGas(WSnm,E,Tmm,Pa,degC)    new energy; for gas-target
    .Eold(   WSnm,E,Tum)            old energy before this target; thick=Tum
    .EoldGas(WSnm,E,Tmm,Pa,degC)    old energy; for gas-target
  < functions(3): Two srWS combination funcs. >
    .Rng_eq_Eold(     WSnm1,WSnm2,E10)        Range equivalent Eold
    .Enew_eq_Th(      WSnm1,WSnm2,E10,Th1)    Enew [MeV/A] eq. Thickness
    .Enew_eq_ThGas(   WSnm1,WSnm2,E10,Th1,Pa1,degC1,Pa2,degC2)   '' for Gas 
    .Enewt_eq_Th(     WSnm1,WSnm2,E10,Th1)    Enew * A [MeV]:total_E eq. Thick
    .Enewt_eq_ThGas(  WSnm1,WSnm2,E10,Th1,Pa1,degC1,Pa2,degC2)   '' for Gas 
    .Enew_eq_Eold(    WSnm1,WSnm2,E10,Th1)    Enew [MeV/A] eq. Eold
    .Enew_eq_EoldGas( WSnm1,WSnm2,E10,Th1,Pa1,degC1,Pa2,degC2)  '' for Gas 
    .Enewt_eq_Eold(   WSnm1,WSnm2,E10,Th1)    Enew * A [MeV]:total_E eq. Eold
    .Enewt_eq_EoldGas(WSnm1,WSnm2,E10,Th1,Pa1,degC1,Pa2,degC2)  '' for Gas 
    .dEu_eq_Th(       WSnm1,WSnm2,E10,Th1)    dEu [MeV/A] equivalent Thick
    .dEu_eq_ThGas(    WSnm1,WSnm2,E10,Th1,Pa1,degC1,Pa2,degC2)  '' for Gas 
    .dEt_eq_Th(       WSnm1,WSnm2,E10,Th1)    dEu  * A [MeV]:total_E eq. Thick
    .dEt_eq_ThGas(    WSnm1,WSnm2,E10,Th1,Pa1,degC1,Pa2,degC2)  '' for Gas 
    .dEu_eq_Eold(     WSnm1,WSnm2,E10,Th1)    dEu [MeV/A] eq. Eold
    .dEu_eq_EoldGas(  WSnm1,WSnm2,E10,Th1,Pa1,degC1,Pa2,degC2)  '' for Gas 
    .dEt_eq_Eold(     WSnm1,WSnm2,E10,Th1)    dEu  * A [MeV]:total_E eq. Eold
    .dEt_eq_EoldGas(  WSnm1,WSnm2,E10,Th1,Pa1,degC1,Pa2,degC2)  '' for Gas 
< functions(9): work-sheet info >
    .WSinfo(     WSnm,modL,*)   Info. of a srWS 
    .WSplotLET(  WSnm,*)        Plot ._LET{e|n|t} of a srWS 
    .WSplotRngSt(WSnm,*)        Plot {._Range|._StLng|._StLtr} of a srimWS 
    .info(       mod, *)        Info. of this srWB
    .WBinfo(          *)        Show Beam vs Target list of this srWB
    .WBdump(          *)        Contents Dump of this srWB to Power-Point
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""

    """ Const.Def.
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    VerInfo = 'pySRIMfit ver.1.1 (2023.03)'
    xlWSid  = 'srim'    # xl WkSht ID
    
    """ class Attributes
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""    
    def _myclear(self):
        #print('@@srimWB:Home dir=',os.path.dirname(__file__))
        self.WBpath = os.path.join(os.path.dirname(__file__),'./_MySRIMwb')
        self.WBname     = 'MySRIMwb.pickle'     # defa. fname
        self.WSL        = []    # [srWS]
        self.nWS        = 0     # len([srWS])
        self.WSnameL    = []    # [srWS.wsName]
        self.WSnameLDB  = []    # [dict{'Bm':*,'Tg':*,..}]
        self.curWS      = None  # srWS
        self.curWSname  = ''    # srWS.wsName
        
    def __init__(self):
        """  Constructor
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        self._myclear()

    """ srWB : load / save
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    
    def loadWS(self, path:str=None, fnm:str=None) -> bool:
        """  load srWB from .pickle
        <optional>
        path:   str     load path (defa. self.WBpath)
        fnm:    str     file name (defa. self.WBname)
        <Ret>   bool    False, if the pickle file not found
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        _path = path if(path is not None) else self.WBpath
        _fnm  = fnm  if(fnm  is not None) else self.WBname
        fn = os.path.join(_path,_fnm)
        ## print('@@srWS.loadWS: fn=',fn)
        if(not os.path.exists(fn)):
            msg= ('#@ loadWS:Err: [{}] not found'.format(fn))
            _Err(1,msg)
            return( False )            
        with open(fn,'rb') as f:
            self.WSL= pickle.load(f)
        self.nWS = len(self.WSL)
        self.WSnameL = [ws.wsName for ws in self.WSL]
        self.WSnameLDB = self.getWSnmLDB(swSort=1)
        self.WBpath = fn
        return( True )

    def saveWS(self, path:str=None, fnm:str=None):
        """  save srWB as .pickle     ->r#101)
        <optional>
        path:   str     load path (defa. self.WBpath)
        fnm:    str     file name (defa. self.WBname)
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        _path = path if(path is not None) else self.WBpath
        _fnm  = fnm  if(fnm  is not None) else self.WBname
        fn = os.path.join(_path,_fnm)
        with open(fn,'wb') as f:
            #pickle.dump(self.WSL,f,pickle.HIGHEST_PROTOCOL)
            pickle.dump(self.WSL,f,protocol=None)
    
    """ srWS : Append / Remove / Sort WSnameL[]
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    
    def _appendWS(self, WS) -> bool:
        """  Append srWS
        WS:     srimWS      to append
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        if(WS.wsName in self.WSnameL):
            msg= ('#@ _appendWS: Err. sheet name=[{}] '
                  'exist already').format(WS.wsName)
            _Err(1,msg)
            return(False)
        self.WSL.append( WS )
        self.WSnameL.append( WS.wsName )
        self.nWS += 1
        return(True)
    
    def _removeWSnm(self, WSnm:str) -> bool:
        """  Remove srWS by its WkSheet name
        WSnm:   str         srimWS name to delete
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        if(WSnm not in self.WSnameL): 
            msg= ('#@ _removeWSnm: Err. sheet name=[{}] ' 
                  'not found.').format(WSnm)
            _Err(1,msg)
            return(False)
        ix = self.WSnameL.index(WSnm)
        self.WSnameL.pop(ix)
        self.WSL.pop(ix)
        self.nWS -= 1
        return(True)

    def _sortWSnmL(self, swSort:int=1):
        """  Sort self.WSnameL[]
        <optional>
        swSort: int     sort mode
                        =1 : sort by Z,A,Tg (defa.)
                        =2 : sort by Tg,Z,A
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        if(swSort not in [1,2]): return
        #print('@@srWB._sortWSnmL: before\n',self.WSnameL)
        wsnLO = self.WSnameL
        WSLO  = self.WSL
        wsnLN = self.getWSnmL(swSort=swSort)
        WSLN = []
        for wsn in wsnLN:
            ix = wsnLO.index(wsn)
            ws = WSLO[ix]
            WSLN.append( ws )
        self.WSL = WSLN
        self.WSnameL = wsnLN
        #print('@@srWB._sortWSnmL: after\n',self.WSnameL)

    """ srWS : getter
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    
    def getWS(self, WSnm:str) -> object:
        """  Find srWS by WSname
        WSnm:   str     work-sheet name as 'srim'+Beam+'_'+Target
        <Ret>   srimWS  or None, if not found
        <Set>   .curWS, .curWSname
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        if(WSnm == self.curWSname):
            return( self.curWS )
        if(WSnm not in self.WSnameL): 
            msg=("#@ getWS: WSnm=[{}] not found"
                 ", then return value = float('NaN') or '' ").format(WSnm)
            _Err(1,msg)
            return(None)
        ix = self.WSnameL.index(WSnm)
        self.curWS = self.WSL[ix]
        self.curWSname = WSnm
        return( self.curWS )

    def getWSL(self, WSnmL:list) -> list:
        """  Find srWS list by WSname list
        WSnmL:  [str]       work-sheet name list
        <Ret>   [srimWS]    or None, if not found
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        wsL = []
        for WSnm in WSnmL:
            if(WSnm not in self.WSnameL): 
                msg= '@@ getWSL: WSnm=[{}] not found'.format(WSnm)
                _Err(1,msg)
                return(None)
            ix = self.WSnameL.index(WSnm)
            wsL.append( self.WSL[ix] )
        return( wsL )

    def getWSnm_AZT(self, ionA:int, ionZ:int, Trg:str, 
                    inWB:bool=False) -> object:
        """  make srWS name by A,Z,Trg / or check it exists in this srWB
        ionA:   int     Beam Mass   number
        ionZ:   int     Beam Atomic number
        Trg:    str     Target name
        <optional>
        inWB:   bool    if True, just check its existance
        <Ret> str|bool  WkSheet name | bool if inWB=True 
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        WSnm = srimWSname_AZT(ionA,ionZ,Trg)
        if(inWB):   return( WSnm in self.WSnameL )
        else:       return( WSnm )

    def getWSnm_ASymT(self, ionA:int, ionSym:str, Trg:str,
                      inWB:bool=False) -> object:
        """  make srWS name by A,Atomic-Symbol,Trg / or check its existance
        ionSym: str     Beam Atomic-symbol; like 'Kr'
        others: -> see) getWSnm_AZT()
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        WSnm = srimWSname_ASymT(ionA,ionSym,Trg)
        if(inWB):   return( WSnm in self.WSnameL )
        else:       return( WSnm )

    def getWSnm_BmT(self, Beam:str, Trg:str, inWB:bool=False) -> object:
        """  make srWS name by Beam,Trg / or check its existance
        Beam:   str     Beam name; like '84Kr'
        others: -> see) getWSnm_AZT()
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        WSnm = srimWSname_BmT(Beam,Trg)
        if(inWB):   return( WSnm in self.WSnameL )
        else:       return( WSnm )    

    def getWSnmL(self, Beam:str=None, Trg:str=None,
                 swSort:int=0 )->list:
        """  get srWS name list
        <optional>
        Beam:   str     Beam   name; like '84Kr', 'Kr'
        Trg:    str     Target name; like 'Air', 'Si'
        swSort: int     sort mode
                        =0 : w/o sort
                        =1 : sort by Z,A,Tg
                        =2 : sort by Tg,Z,A
        <Ret> list of str; like ..
            ['srim84Kr_Air', ..]
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        SpL = self.getWSnmLDB(Beam,Trg,swSort)
        wsnL = [d.get('WSnm') for d in SpL]
        return( wsnL )

    def getWSnmLDB(self,Beam:str=None,Trg:str=None,
                 swSort:int=0 )->list:
        """  get srWS name list as dict DB
        <optional>
          see) .getWSnmL()
        ret) list of dict which are matched; like ..
            [{'Bm':'84Kr','A':84,'Sym':'Kr','Z':36,  
              'Tg':'Air', 'WSnm':'srim84Kr_Air'}, ..]
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        wsnL = self.WSnameL
        SpL = []            # WSname splited list
        for wsn in wsnL:
            SpL.append( srimWSname_split(wsn) )
        # print(wsnSpL)
        SpL1 = (SpL  if(Beam is None) else 
                [v for v in SpL  if Beam in v['Bm']] )
        SpL2 = (SpL1 if(Trg  is None) else 
                [v for v in SpL1 if Trg  in v['Tg']] )
        if(swSort == 1):              # ->r#201)
            SpL3 = sorted(SpL2,key=operator.itemgetter('Z','A','Tg'))
            return( SpL3 )
        elif(swSort == 2):
            SpL3 = sorted(SpL2,key=operator.itemgetter('Tg','Z','A'))
            return( SpL3 )
        else:            
            return( SpL2 )

    """ srWS : Attributes getter
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""

    def wsName(self, WSnm:str) -> str:
        """ <Ret>  WSnm .wsName """
        ws = self.getWS(WSnm)
        return( "" if(ws is None) else ws.wsName )
    
    def SRIMver(self, WSnm:str) -> str:
        """ <Ret>  WSnm .SRIMver """
        ws = self.getWS(WSnm)
        return( "" if(ws is None) else ws.SRIMver )
    
    def IonZ(self, WSnm:str) -> int:
        """ <Ret>  WSnm .IonZ """
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.IonZ )
    
    def IonA(self, WSnm:str) -> int:
        """ <Ret>  WSnm .IonA """
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.IonA )
    
    def Beam(self, WSnm:str) -> str:
        """ <Ret>  WSnm .Beam """
        ws = self.getWS(WSnm)
        return( "" if(ws is None) else ws.Beam )
    
    def Target(self, WSnm:str) -> str:
        """ <Ret>  WSnm .Target """
        ws = self.getWS(WSnm)
        return( "" if(ws is None) else ws.Target )
    
    def TargetNm(self, WSnm:str) -> str:
        """ <Ret>  WSnm .TargetNm """
        ws = self.getWS(WSnm)
        return( "" if(ws is None) else ws.TargetNm )
    
    def TrgDens(self, WSnm:str) -> float:
        """ <Ret>  WSnm .TrgDens [g/cm3] """
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.TrgDens )
    
    def TrgDensAtm(self, WSnm:str) -> float:
        """ <Ret>  WSnm .TrgDensAtm [atoms/cm3] """
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.TrgDensAtm )
    
    def TrgCmp(self, WSnm:str) -> list:
        """ <Ret>  WSnm ._TrgCmp : Target material composition table """
        ws = self.getWS(WSnm)
        return( [] if(ws is None) else ws.TrgCmpL() )
    
    def BraggC(self, WSnm:str) -> float:
        """ <Ret>  WSnm .BraggC : Bragg corection [float] not [%] """
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.BraggC )
    
    def Gas(self, WSnm:str) -> bool:
        """ <Ret>  WSnm .Gas : Target is Gas ? [bool] """
        ws = self.getWS(WSnm)
        return( False if(ws is None) else ws.Gas )
    
    def GasPtbl(self, WSnm:str) -> float:
        """ <Ret>  WSnm .GasPtbl : Gas Press.[Pa] asuumed in srWS """
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.GasPtbl )
    
    def GasTtbl(self, WSnm:str) -> float:
        """ <Ret>  WSnm .GasTtbl : Gas Temp.[degC] asuumed in srWS """
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.GasTtbl )
    
    def Corded(self, WSnm:str) -> str:
        """ <Ret>  WSnm .Corded : Corded by info. """
        ws = self.getWS(WSnm)
        return( "" if(ws is None) else ws.Corded )
    
    def minE(self, WSnm:str) -> float:
        """ <Ret>  WSnm .minE : min of E table """
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.minE )
    
    def maxE(self, WSnm:str) -> float:
        """ <Ret>  WSnm .maxE : max of E table """
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.maxE )
    
    def minLETe(self, WSnm:str, Uid:int=0) -> float:
        """ <Ret>  WSnm .minLETe : min of LET_elec table in Uid unit """
        ws = self.getWS(WSnm)
        if(ws is None):     return( float('NaN') )
        let = ws.minLETe
        if(Uid != 0):    let *= ws.CnvF(Uid)
        return( let )
    
    def maxLETe(self, WSnm:str, Uid:int=0) -> float:
        """ <Ret>  WSnm .maxLETe : max of LET_elec table in Uid unit """
        ws = self.getWS(WSnm)
        if(ws is None):     return( float('NaN') )
        let = ws.maxLETe
        if(Uid != 0):    let *= ws.CnvF(Uid)
        return( let )
    
    def minLETn(self, WSnm:str, Uid:int=0) -> float:
        """ <Ret>  WSnm .minLETn : min of LET_nucl table in Uid unit """
        ws = self.getWS(WSnm)
        if(ws is None):     return( float('NaN') )
        let = ws.minLETn
        if(Uid != 0):    let *= ws.CnvF(Uid)
        return( let )
    
    def maxLETn(self, WSnm:str, Uid:int=0) -> float:
        """ <Ret>  WSnm .maxLETn : max of LET_nucl table in Uid unit """
        ws = self.getWS(WSnm)
        if(ws is None):     return( float('NaN') )
        let = ws.maxLETn
        if(Uid != 0):    let *= ws.CnvF(Uid)
        return( let )
    
    def minLETt(self, WSnm:str, Uid:int=0) -> float:
        """ <Ret>  WSnm .minLETt : min of LET_tot table in Uid unit """
        ws = self.getWS(WSnm)
        if(ws is None):     return( float('NaN') )
        let = ws.minLETt
        if(Uid != 0):    let *= ws.CnvF(Uid)
        return( let )
    
    def maxLETt(self, WSnm:str, Uid:int=0) -> float:
        """ <Ret>  WSnm .maxLETt : max of LET_tot table in Uid unit """
        ws = self.getWS(WSnm)
        if(ws is None):     return( float('NaN') )
        let = ws.maxLETt
        if(Uid != 0):    let *= ws.CnvF(Uid)
        return( let )
    
    def minRng(self, WSnm:str) -> float:
        """ <Ret>  WSnm .minRng : min of Range tabl """
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.minRng )
    
    def maxRng(self, WSnm:str) -> float:
        """ <Ret>  WSnm .maxRng : max of Range tabl """
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.maxRng )
    
    def minStLng(self, WSnm:str) -> float:
        """ <Ret>  WSnm .minStLng : min of Straggling Longitudinal table """
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.minStLng )
    
    def maxStLng(self, WSnm:str) -> float:
        """ <Ret>  WSnm .maxStLng : max of Straggling Longitudinal table """
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.maxStLng )
    
    def minStLtr(self, WSnm:str) -> float:
        """ <Ret>  WSnm .minStLtr : min of Straggling Lateral table """
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.minStLtr )
    
    def maxStLtr(self, WSnm:str) -> float:
        """ <Ret>  WSnm .maxStLtr : max of Straggling Lateral table """
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.maxStLtr )
    
    def maxLETe2E(self, WSnm:str) -> float:
        """ <Ret>  WSnm .maxLETe2E : E at max LET_elec """
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.maxLETe2E )
    
    def maxLETn2E(self, WSnm:str) -> float:
        """ <Ret>  WSnm .maxLETn2E : E at max LET_nucl """
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.maxLETenE )
    
    def maxLETt2E(self, WSnm:str) -> float:
        """ <Ret>  WSnm .maxLETt2E : E at max LET_tot """
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.maxLETt2E )

    """ srWS : unit conversion
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""

    def um2mg(self, WSnm:str, Tum:float) -> float:
        """ Thickness convertion : micron -> mg
        Tum :       float   [um] thickness in micron
        <Ret> Tmg   float   [mg] = um * .TrgDens / 10
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.um2mg(Tum) )
       
    def mg2um(self, WSnm:str, Tmg:float) -> float:
        """ Thickness convertion : mg -> micron
        Tmg :       float   [mg] thickness in mg
        <Ret> Tum   float   [um] = mg * 10 / .TrgDens
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.mg2um(Tmg) )

    def mm2umGas(self, WSnm:str,Tmm:float, Pa:float, degC:float) -> float:
        """ Gas Thickness convertion : mm -> micron
        Tmm :       float   [mm] thickness in mm
        Pa:         float   [Pa]    Gas pressure
        degC:       float   [degC]  Gas temperature
        <Ret> Tum   float   [um] = Tmm * 1000 * ThkStd(Pa,degC)
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.mm2umGas(Tmm,Pa,degC) )
    
    def um2mmGas(self, WSnm:str,Tum:float, Pa:float, degC:float) -> float:
        """ Gas Thickness convertion : mm -> micron
        Tum :       float   [um] thickness in micron
        Pa:         float   [Pa]    Gas pressure
        degC:       float   [degC]  Gas temperature
        <Ret> Tmm   float   [mm] = Tum / 1000 / ThkStd(Pa,degC)
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.um2mmGas(Tum,Pa,degC) )

    def LETunit(self, Uid:int) -> str:
        """ <Ret> str LET unit name """
        ok = True if(Uid in srimWS_Uid) else False
        if(not ok):
            msg = '#@ LETunit:Err: Uid=[{}] invalid'.format(Uid)
            _Err(1,msg)
            return( "" )
        return( srimWS_Uid[Uid] )
    
    def LETcnvF(self, WSnm:str, Uid:int) -> float:
        """ <Ret> float LET unit conversion factor : LET*=CnvF """
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.CnvF(Uid) )
    
    """ srWS : E --> Value
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    
    def LETe(self, WSnm:str, E:float, Uid:int=0) -> float:
        """ LETe(E) dE/dX electric part
        <Ret>  LETe(E) * CnvF(Uid) or 0 (if Uid error)
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.LETe(E,Uid) )
    
    def LETn(self, WSnm:str, E:float, Uid:int=0) -> float:
        """ LETn(E) dE/dX nuclear part
        <Ret>  LETn(E) * CnvF(Uid) or 0 (if Uid error)
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.LETn(E,Uid) )
    
    def LETt(self, WSnm:str, E:float, Uid:int=0) -> float:
        """ LETt(E) dE/dX total= electric + nuclear part
        <Ret>  LETt(E) * CnvF(Uid) or 0 (if Uid error)
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.LETt(E,Uid) )
    
    def Rng(self, WSnm:str, E:float) -> float:
        """ <Ret>  Rng(E) [um] Range """
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.Rng(E) )
    
    def StLng(self, WSnm:str, E:float) -> float:
        """ <Ret>  StLng(E) [um] Straggling Longitudinal """
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.StLng(E) )
    
    def StLtr(self, WSnm:str, E:float) -> float:
        """ <Ret>  StLng(E) [um] Straggling Lateral """
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.StLtr(E) )
    
    def E2LETe(self, WSnm:str, E:float, Uid:int=0) -> float:
        """ == LETe(E,Uid) """
        return( self.LETe(WSnm,E,Uid) )
    
    def E2LETn(self, WSnm:str, E:float, Uid:int=0) -> float:
        """ == LETn(E,Uid) """
        return( self.LETn(WSnm,E,Uid) )
    
    def E2LETt(self, WSnm:str, E:float, Uid:int=0) -> float:
        """ == LETt(E,Uid) """
        return( self.LETt(WSnm,E,Uid) )
    
    def E2Rng(self, WSnm:str, E:float) -> float:
        """ == Rng(E) """
        return( self.Rng(WSnm,E) )
    
    def E2StLng(self, WSnm:str, E:float) -> float:
        """ == StLng(E) """
        return( self.StLng(WSnm,E) )
    
    def E2StLtr(self, WSnm:str, E:float) -> float:
        """ == StLtr(E) """
        return( self.StLtr(WSnm,E) )
    
    def LETe2E(self, WSnm:str, LETe:float, *,
               Uid:int=0, Hside:bool=True) -> float:
        """  LETe --> E [MeV/A] : inv.func. of LETe(E)
        LETe:   float   [Uid unit]
        <optional: kwd args>
        Uid:    int     Unit ID#  (defa.=0)
        Hside:  bool    if(True) search at E-higher side of the LETe peak
                        else      ''       E-lower-side  ''
        <Ret>  E  float   [MeV/A]
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) 
               else ws.LETe2E(LETe,Uid=Uid,Hside=Hside) )
    
    def LETn2E(self, WSnm:str, LETn:float, *,
               Uid:int=0, Hside:bool=True) -> float:
        """  LETn --> E [MeV/A] : inv.func. of LETn(E)
        see) LETe2E() for detail
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) 
               else ws.LETn2E(LETn,Uid=Uid,Hside=Hside) )
    
    def LETt2E(self, WSnm:str, LETt:float, *,
               Uid:int=0, Hside:bool=True) -> float:
        """  LETt --> E [MeV/A] : inv.func. of LETt(E)
        see) LETe2E() for detail
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) 
               else ws.LETt2E(LETt,Uid=Uid,Hside=Hside) )
    
    def Rng2E(self, WSnm:str, Rng:float) -> float:
        """  Rng --> E [MeV/A] : inv.func. of Rng(E)  """
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.Rng2E(Rng) )
    
    def StLng2E(self, WSnm:str, Str:float) -> float:
        """  StLng --> E [MeV/A] : inv.func. of StLng(E)  """
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.StLng2E(Str) )
    
    def StLtr2E(self, WSnm:str, Str:float) -> float:
        """  StLtr --> E [MeV/A] : inv.func. of StLtr(E)  """
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.StLtr2E(Str) )
    
    def Enew(self, WSnm:str, E0:float, Tum:float) -> float:
        """  E0 -> Thick[um] -> E1:new-energy
        E0:     float   [MeV/A] Beam energy
        Tum:    float   [um]    Target thickness 'in micron'
        <Ret>
        E1:     float   [MeV/A] Beam new-energy through the Tum
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.Enew(E0,Tum) )
    
    def EnewGas(self, WSnm:str, E0:float, Tmm:float,
                Pa:float, degC:float) -> float:
        """  E0 -> Thick[um],Pa,degC -> E1:new-energy
        E0:     float   [MeV/A] Beam energy
        Tmm:    float   [mm]    Target thickness 'in mm'
        Pa:     float   [Pa]    Gas pressure
        degC:   float   [degC]  Gas temperature
        <Ret>
        E1:     float   [MeV/A] Beam new-energy through the Tmm
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.EnewGas(E0,Tmm,Pa,degC) )
    
    def Eold(self, WSnm:str, E1:float, Tum:float) -> float:
        """  E0:old-energy <- Thick[um] <- E1
        E1:     float   [MeV/A] Beam energy
        Tum:    float   [um]    Target thickness 'in micron'
        <Ret>
        E0:     float   [MeV/A] Beam old-energy before the Tum
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.Eold(E1,Tum) )
    
    def EoldGas(self, WSnm:str, E1:float, Tmm:float,
                Pa:float, degC:float) -> float:
        """  E0:old-energy <- Thick[um],Pa,degC <- E1
        E0:     float   [MeV/A] Beam energy
        Tmm:    float   [mm]    Target thickness 'in mm'
        Pa:     float   [Pa]    Gas pressure
        degC:   float   [degC]  Gas temperature
        <Ret>  
        E0:     float   [MeV/A] Beam old-energy before the Tmm
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.EoldGas(E1,Tmm,Pa,degC) )
    
    def ThkStdGas(self, WSnm:str, Pa:float, degC:float) -> float:
        """ Gas thick. conv. factor to "Standard P,T" of this SRIM table
        Pa:     float   [Pa]    Gas pressure
        degC:   float   [degC]  Gas temperature
        <Ret>  
        CFct:   float   [1]     Conversion factor
        (usage) SRIMtbl_Thick[um] = Thick[um]         * CFct
                        Thick[um] = SRIMtbl_Thick[um] / CFct
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        ws = self.getWS(WSnm)
        return( float('NaN') if(ws is None) else ws.ThkStdGas(Pa,degC) )
    
    """ Two srWS combination func. : Solid & Gas
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""

    """ DEBUG messages"""

    def _msgW_E10_lt0(self, fnm:str, E10:float, inf:str):
        ms1= '#@ {}:Warn E10=[{:.2f}] <= 0 '.format(fnm,E10)
        ms2= ' ; then {} becomes indefinite.'.format(inf)
        return( [ms1,ms2] )

    def _msgW_Enew1_lt0(self, fnm:str, wsnm:str, 
                        E10:float, Th1:float, E11:float, inf:str):
        ms1= '#@ {}:Warn:'.format(fnm)
        ms2= self._msgW_Enew(wsnm,'E10','Th1',E10,Th1,E11) 
        ms3= ' = E11 <= 0 stopped'
        ms4= ' ; then {} becomes indefinite.'.format(inf)
        return( [ms1,ms2+ms3,ms4] )

    def _msgW_Enew(self, wsnm:str, Enm:str, Thnm:str, 
                  E0:float, Th:float, E1:float):
        ms1= ' Enew([{}],{}[{:.2e}],{}[{:.2e}])= [{:.2e}]'.format(
                wsnm,Enm,E0,Thnm,Th,E1)
        return( ms1 )
    
    def _msgW_Th_lt0(self, fnm:str, Thnm:str, Th:float):
        ms1= '#@ {}:Warn:'.format(fnm)
        ms2= ' {}[{:.2e}] <= 0 invalid.'.format(Thnm,Th) 
        return( ms1+ms2 )        

    """ Error check for Gas """

    def _Gas_eq_Gas(self, msgH:str, WSnm1:str, WSnm2:str,
            Pa1:float,degC1:float,Pa2:float,degC2:float) -> tuple:
        """ check Gas1, Gas2 & return conv. factors 
        msgH:   str     _Err() message header = caller func. name
        WSnm1:  str     srWS name for E10, Th1
        WSnm2:  str     srWS name for E20, Th2
        Pa1:    float   [Pa]    Gas pressure     for WS1
        degC1:  float   [degC]  Gas temperature  for WS1
        Pa2:    float   [Pa]    Gas pressure     for WS2
        degC2:  float   [degC]  Gas temperature  for WS2
        <Ret>           conv. factors for (Gas1,Gas2)
        kG1:    float   use for Th1 [um] = Tmm1 [mm] * kG1
        kG2:    float   use for Tmm2[mm] = Th2  [um] / kG2
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        ws1,ws2,ms = self.getWS(WSnm1), self.getWS(WSnm2), ''

        def _chk_gas(Nm,ws,wsnm,Pa,degC):
            global ms
            if(not ws.Gas):
                ms= '#@ {}:Warn: WSnm{}[{}] is NOT Gas'.format(msgH,Nm,wsnm)
                return(False)
            if(Pa < 0):
                ms= '#@ {}:Warn: Pa{}[{:.2e}] <= 0 invalid'.format(msgH,Nm,Pa)
                return(False)
            if(degC < -273.15):
                ms= '#@ {}:Warn: degC{}[{:.2f}] <= -273.15 invalid'.format(
                    msgH,Nm,degC)
                return(False)
            return(True)

        if(not _chk_gas('1',ws1,WSnm1,Pa1,degC1)):
            _Err(1,ms)
            return(-1,-1)
        if(not _chk_gas('2',ws2,WSnm2,Pa2,degC2)):
            _Err(1,ms)
            return(-1,-1)
        kG1= ws1.ThkStdGas(Pa1,degC1) * 1000.               
        kG2= ws2.ThkStdGas(Pa2,degC2) * 1000.
        return( kG1, kG2 )               

    """ Two srWS combination func. : Rng_eq_Eold(): Gas is same
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""

    def Rng_eq_Eold(self, WSnm1:str, WSnm2:str,
                  E10:float ) -> float:
        """ Range equivalent Eold
            ie.) get E20 [MeV/A] of 
                     Rng(WS1,E10) == Rng(WS2,E20)
        WSnm1:  str     srWS name for E10
        WSnm2:  str     srWS name for E20
        E10:    float   [MeV/A] Beam energy for WS1
        <Ret> 
        E20:    float   [MeV/A] Range equivalent Eold for WS2
            = 0     if (E10 <= 0)           
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        ws1,ws2 = self.getWS(WSnm1), self.getWS(WSnm2)
        if(E10 <= 0.):
            return( 0. )
        R10 = ws1.Rng(E10)      # Warn: R10 out of ws1._Rng ; checked by ws.Rng()
        R20 = R10
        E20 = ws2.Rng2E(R20)    # Warn: R20 out of ws2._Rng ; ''
        return( E20 )

    """ Two srWS combination func. : Enew_eq_Th() Enew_eq_ThGas()
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""

    def Enew_eq_Th(self, WSnm1:str, WSnm2:str,
                 E10:float, Th1:float ) -> float:
        """ Enew [MeV/A] equivalent Thickness
            ie.) get Th2 [um] of
                     Enew(WS1,E10       ,Th1) ; E11 
                  == Enew(WS2,E20 == E10,Th2) ; E21
        WSnm1:  str     srWS name for E10, Th1
        WSnm2:  str     srWS name for E20, Th2
        E10:    float   [MeV/A] Beam energy   for WS1
        Th1:    float   [um]    Target thick. for WS1
        <Ret> 
        Th2:    float   [um]    Enew equiv. thick. for WS2
            = NaN   if (E10 <= 0) | (E11 <= 0) | (Th1 < 0)       
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        Th2 = self._Enew_eq_Th('Enew_eq_Th',WSnm1,WSnm2,E10,Th1)
        return( Th2 )

    def Enew_eq_ThGas(self, WSnm1:str, WSnm2:str,
            E10:float, Tmm1:float, 
            Pa1:float, degC1:float, Pa2:float, degC2:float ) -> float:
        """ Enew [MeV/A] equivalent Thickness : Gas
            ie.) get Tmm2 of
                     EnewGas(WS1,E10,       Tmm1,Pa1,degC1) ; E11
                  == EnewGas(WS2,E20 == E10,Tmm2,Pa2,degC2) ; E21
        WSnm1:  str     srWS name for E10, Th1
        WSnm2:  str     srWS name for E20, Th2
        E10:    float   [MeV/A] Beam energy      for WS1
        Tmm1:   float   [mm]    Target thickness for WS1
        Pa1:    float   [Pa]    Gas pressure     for WS1
        degC1:  float   [degC]  Gas temperature  for WS1
        Pa2:    float   [Pa]    Gas pressure     for WS2
        degC2:  float   [degC]  Gas temperature  for WS2
        <Ret> 
        Tmm2:   float   [mm]    Enew equiv. thick. for WS2
            = NaN   if (E10 <= 0) | (E11 <= 0) | (Th1 < 0)       
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        msgH = 'Enew_eq_ThGas'
        kG1,kG2 = self._Gas_eq_Gas(msgH,WSnm1,WSnm2,Pa1,degC1,Pa2,degC2)             
        Th1 = Tmm1 * kG1
        Th2 = self._Enew_eq_Th(msgH,WSnm1,WSnm2,E10,Th1) 
        Tmm2= Th2 / kG2
        return( Tmm2 )

    def _Enew_eq_Th(self, msgH:str, WSnm1:str, WSnm2:str,
            E10:float, Th1:float) -> float:
        """ Enew [MeV/A] equivalent Thickness : common
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        ws1,ws2 = self.getWS(WSnm1), self.getWS(WSnm2)
        if(E10 <= 0.):
            msgL= self._msgW_E10_lt0(msgH,E10,'Th2')
            _Err(1,'\n'.join(msgL))
            return( float('NaN') )
        if(Th1 < 0):                # Th1 = 0 is valid
            msg = self._msgW_Th_lt0(msgH,'Th1',Th1)
            _Err(1,msg)
            return( float('NaN') )
        E11= ws1._Enew(E10,Th1)     # use _Enew() to avoid NOT Gas msg.
        if(E11 <= 0.):
            msgL= self._msgW_Enew1_lt0(msgH,WSnm1,E10,Th1,E11,'Th2')
            _Err(1,'\n'.join(msgL))
            return( float('NaN') )
        E20 = E10
        R20 = ws2.Rng(E20)  # Warn: R20 out of ws2._Rng ; checked by ws.Rng()
        E21 = E11
        R21 = ws2.Rng(E21)  # Warn: R21 out of ws2._Rng ; ''
        Th2 = R20 - R21
        return( Th2 )
    
    """ Two srWS combination func. : Enewt_eq_Th() Enewt_eq_ThGas()
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""

    def Enewt_eq_Th(self, WSnm1:str, WSnm2:str,
                 E10:float, Th1:float) -> float:
        """ Enew * A [MeV]: total_E equivalent Thickness
            ie.) get Th2 [um] of 
                   Enew(WS1,E10,       Th1) * A1 ; E11*A1
                == Enew(WS2,E20 == E10,Th2) * A2　; E21*A2
        WSnm1:  str     srWS name for E10, Th1
        WSnm2:  str     srWS name for E20, Th2
        E10:    float   [MeV/A] Beam total-energy for WS1
        Th1:    float   [um]    Target thick. for WS1
        <Ret> 
        Th2:    float   [um]    Enew_total [MeV] equiv. Thick. for WS2
            = NaN   if (E10 <= 0) | (E11 <= 0) | (E11*A1 > E20*A2) | (Th1 < 0)         
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        Th2 = self._Enewt_eq_Th('Enewt_eq_Th',WSnm1,WSnm2,E10,Th1)
        return( Th2 )

    def Enewt_eq_ThGas(self, WSnm1:str, WSnm2:str,
            E10:float, Tmm1:float, 
            Pa1:float, degC1:float, Pa2:float, degC2:float ) -> float:
        """ Enew * A [MeV]: total_E equivalent Thickness : Gas
            ie.) get Tmm2 of 
                     EnewGas(WS1,E10,       Tmm1,Pa1,degC1) * A1 ; E11*A1 
                  == EnewGas(WS2,E20 == E10,Tmm2,Pa2,degC2) * A2 ; E21*A2
        WSnm1:  str     srWS name for E10, Th1
        WSnm2:  str     srWS name for E20, Th2
        E10:    float   [MeV/A] Beam total-energy for WS1
        Tmm1:   float   [mm]    Target thickness  for WS1
        Pa1:    float   [Pa]    Gas pressure      for WS1
        degC1:  float   [degC]  Gas temperature   for WS1
        Pa2:    float   [Pa]    Gas pressure      for WS2
        degC2:  float   [degC]  Gas temperature   for WS2
        Tmm2:   float   [mm]    Enew_total [MeV] equiv. Thick. for WS2
            = NaN   if (E10 <= 0) | (E11 <= 0) | (E11*A1 > E20*A2) | (Th1 < 0)         
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        msgH = 'Enewt_eq_ThGas'
        kG1,kG2 = self._Gas_eq_Gas(msgH,WSnm1,WSnm2,Pa1,degC1,Pa2,degC2)             
        Th1 = Tmm1 * kG1
        Th2 = self._Enewt_eq_Th(msgH,WSnm1,WSnm2,E10,Th1) 
        Tmm2= Th2 / kG2
        return( Tmm2 )

    def _Enewt_eq_Th(self, msgH:str, WSnm1:str, WSnm2:str,
            E10:float, Th1:float) -> float:
        """ Enew * A [MeV]: total_E equivalent Thickness : common 
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        ws1,ws2 = self.getWS(WSnm1), self.getWS(WSnm2)
        A1,A2   = ws1.IonA, ws2.IonA
        if(E10 <= 0.):
            msgL= self._msgW_E10_lt0(msgH,E10,'Th2')
            _Err(1,'\n'.join(msgL))
            return( float('NaN') )
        if(Th1 < 0):                # Th1 = 0 is valid
            msg = self._msgW_Th_lt0(msgH,'Th1',Th1)
            _Err(1,msg)
            return( float('NaN') )
        E11= ws1._Enew(E10,Th1)     # use _Enew() to avoid NOT Gas msg.
        if(E11 <= 0.):
            msgL= self._msgW_Enew1_lt0(msgH,WSnm1,E10,Th1,E11,'Th2')
            _Err(1,'\n'.join(msgL))
            return( float('NaN') )
        E11t = E11*A1
        E20  = E10
        if(E11t > E20*A2):
            ms1= '#@ {}:Warn:'.format(msgH) 
            ms2= self._msgW_Enew(WSnm1,'E10','Th1',E10,Th1,E11)+' = E11'
            ms3= '   E11[{:.2f}] * A1[{:3d}] = [{:.2f}] MeV'.format(E11,A1,E11t)
            ms4= ' > E20[{:.2f}] * A2[{:3d}] = [{:.2f}] MeV'.format(E20,A2,E20*A2)
            ms5= ' ; then there is no answer.'
            _Err(1,'\n'.join([ms1,ms2,ms3,ms4,ms5]))
            return( float('NaN') )
        R20  = ws2.Rng(E20)     # Warn: R20 out of ws2._Rng ; checked by ws.Rng()
        E21t = E11t
        E21  = E21t/A2
        R21  = ws2.Rng(E21)     # Warn: R21 out of ws2._Rng ; ''
        Th2  = R20 - R21
        return( Th2 )

    """ Two srWS combination func. : Enew_eq_Eold() Enew_eq_EoldGas()
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    
    def Enew_eq_Eold(self, WSnm1:str, WSnm2:str,
                     E10:float, Th1:float ) -> float:
        """ Enew [MeV/A] equivalent Eold
            ie.) get E20 [MeV/A] of
                     Enew(WS1,E10,Th1)          ; E11
                  == Enew(WS2,E20,Th2 == Th1)   ; E21
        WSnm1:  str     srWS name for E10, Th1
        WSnm2:  str     srWS name for E20, Th2
        E10:    float   [MeV/A] Beam energy   for WS1
        Th1:    float   [um]    Target thick. for WS1
        <Ret> 
        E20:    float   [MeV/A] Enew equiv. Eold for WS2
            = NaN   if (E10 <= 0) | (Th1 < 0)    
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        Th2 = Th1
        E20 = self._Enew_eq_Eold('Enew_eq_Eold',WSnm1,WSnm2,E10,Th1,Th2)
        return( E20 )

    def Enew_eq_EoldGas(self, WSnm1:str, WSnm2:str,
            E10:float, Tmm1:float,
            Pa1:float, degC1:float, Pa2:float, degC2:float ) -> float:
        """ Enew [MeV/A] equivalent Eold : Gas
            ie.) get E20 of
                     EnewGas(WS1,E10,Tmm1,        Pa1,degC1) ; E11
                  == EnewGas(WS2,E20,Tmm2 == Tmm1,Pa2,degC2) ; E21
        WSnm1:  str     srWS name for E10, Th1
        WSnm2:  str     srWS name for E20, Th2
        E10:    float   [MeV/A] Beam energy      for WS1
        Tmm1:   float   [mm]    Target thickness for WS1
        Pa1:    float   [Pa]    Gas pressure     for WS1
        degC1:  float   [degC]  Gas temperature  for WS1
        Pa2:    float   [Pa]    Gas pressure     for WS2
        degC2:  float   [degC]  Gas temperature  for WS2
        <Ret> 
        E20:    float   [MeV/A] Enew equiv. Eold for WS2
            = NaN   if (E10 <= 0) | (Th1 < 0)    
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        msgH = 'Enew_eq_EoldGas'
        kG1,kG2 = self._Gas_eq_Gas(msgH,WSnm1,WSnm2,Pa1,degC1,Pa2,degC2)             
        Th1 = Tmm1 * kG1
        Th2 = Tmm1 * kG2
        E20 = self._Enew_eq_Eold(msgH,WSnm1,WSnm2,E10,Th1,Th2) 
        return( E20 )
    
    def _Enew_eq_Eold(self, msgH:str, WSnm1:str, WSnm2:str,
            E10:float, Th1:float, Th2:float) -> float:
        """ Enew equivalent Eold : common 
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        ws1,ws2 = self.getWS(WSnm1), self.getWS(WSnm2)
        if(E10 <= 0.):
            msgL= self._msgW_E10_lt0(msgH,E10,'Eold:E20')
            _Err(1,'\n'.join(msgL))
            return( float('NaN') )
        if(Th1 < 0):                 # Th1 = 0 is valid
            msg = self._msgW_Th_lt0(msgH,'Th1',Th1)
            _Err(1,msg)
            return( float('NaN') )
        E11= ws1._Enew(E10,Th1)
        if(E11 <= 0.):              # use _Enew() to avoid NOT Gas msg.
            msgL= self._msgW_Enew1_lt0(msgH,WSnm1,E10,Th1,E11,'Th2')
            _Err(1,'\n'.join(msgL))
            return( float('NaN') )            
        E21 = E11
        #if(E21 <= 0.):  E20 = ws2.Run2E(Th2)
        #else:           E20 = ws2.Eold(E21,Th2)
        E20= ws2._Eold(E21,Th2)     # use _Eold() to avoid NOT Gas msg.
        return( E20 )

    """ Two srWS combination func. : Enewt_eq_Th() Enewt_eq_ThGas()
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""

    def Enewt_eq_Eold(self, WSnm1:str, WSnm2:str,
                 E10:float, Th1:float ) -> float:
        """ Enew * A [MeV]: total_E equivalent Eold
            ie.) get E20 [MeV/A] of 
                     Enew(WS1,E10,Th1)          ; E11 * A1
                 ==  Enew(WS2,E20,Th2 == Th1)   ; E21 * A2
        WSnm1:  str     srWS name for E10, Th1
        WSnm2:  str     srWS name for E20, Th2
        E10:    float   [MeV/A] Beam energy   for WS1
        Th1:    float   [um]    Target thick. for WS1
        <Ret> 
        E20:    float   [MeV/A] Enew_total [MeV] equiv. Eold for WS2
            = NaN   if (E10 <= 0) | (Th1 < 0)    
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        Th2 = Th1
        E20 = self._Enewt_eq_Eold('Enewt_eq_Eold',WSnm1,WSnm2,E10,Th1,Th2) 
        return( E20 ) 

    def Enewt_eq_EoldGas(self, WSnm1:str, WSnm2:str,
                E10:float, Tmm1:float, 
                Pa1:float, degC1:float, Pa2:float, degC2:float ) -> float:
        """ Enew * A [MeV]: total_E equivalent Eold : Gas
            ie.) get E20 [MeV/A] of
                     EnewGas(WS1,E10,Tmm1,        Pa1,degC1) ; E11 * A1
                  == EnewGas(WS2,E20,Tmm2 == Tmm1,Pa2,degC2) ; E21 * A2
        WS1:    str     srWS name for E10, Th1
        WS2:    str     srWS name for E20, Th2
        E10:    float   [MeV/A] Beam energy      for WS1
        Tmm1:   float   [mm]    Target thickness for WS1
        Pa1:    float   [Pa]    Gas pressure     for WS1
        degC1:  float   [degC]  Gas temperature  for WS1
        Pa2:    float   [Pa]    Gas pressure     for WS2
        degC2:  float   [degC]  Gas temperature  for WS2
        <Ret> 
        E20:    float   [MeV/A] Enew_total [MeV] equiv. Eold for WS2
            = NaN   if (E10 <= 0) | (Th1 < 0)    
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        msgH = 'Enewt_eq_EoldGas'
        kG1,kG2 = self._Gas_eq_Gas(msgH,WSnm1,WSnm2,Pa1,degC1,Pa2,degC2)             
        Th1 = Tmm1 * kG1
        Th2 = Tmm1 * kG2
        E20 = self._Enewt_eq_Eold(msgH,WSnm1,WSnm2,E10,Th1,Th2) 
        return( E20 )

    def _Enewt_eq_Eold(self, msgH:str, WSnm1:str, WSnm2:str,
            E10:float, Th1:float, Th2:float) -> float:
        """ Enew * A [MeV]: total_E equivalent Eold : common
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        ws1,ws2 = self.getWS(WSnm1), self.getWS(WSnm2)
        A1,A2   = ws1.IonA, ws2.IonA
        if(E10 <= 0.):
            msgL= self._msgW_E10_lt0(msgH,E10,'Eold')
            _Err(1,'\n'.join(msgL))
            return( float('NaN') )
        if(Th1 < 0):                 # Th1 = 0 is valid
            msg = self._msgW_Th_lt0(msgH,'Th1',Th1)
            _Err(1,msg)
            return( float('NaN') )
        E11= ws1._Enew(E10,Th1)      # use _Enew() to avoid NOT Gas msg.
        if(E11 <= 0.):
            msgL= self._msgW_Enew1_lt0(msgH,WSnm1,E10,Th1,E11,'Eold')
            _Err(1,'\n'.join(msgL))
            return( float('NaN') )
        E11t= E11*A1
        E21t= E11t
        E21 = E21t / A2
        #if(E21t <= 0.): E20 = ws2.Run2E(Th2)
        #else:           E20 = ws2.Eold(E21,Th2)
        E20= ws2._Eold(E21,Th2)     # use _Eold() to avoid NOT Gas msg.
        return( E20 )

    """ Two srWS combination func. : dEu_eq_Th() dEu_eq_ThGas()
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""

    def dEu_eq_Th(self, WSnm1:str, WSnm2:str,
                 E10:float, Th1:float ) -> float:
        """ dEu [MeV/A] equivalent Thickness
            ie.) get Th2 [um] of 
                dE1 = (E10        - Enew(WS1,E10,Th1); E11 ) 
             == dE2 = (E20 == E10 - Enew(WS2,E20,Th2); E21 )
            ie.) get Th2 of Enew(WS1,E10,Th1) = Enew(WS2,E20==E10,Th2)
            ie.) this is identical with Enew_eq_Th(WSnm1,WSnm2,E10,Th1)
        WSnm1:  str     srWS name for E10, Th1
        WSnm2:  str     srWS name for E20, Th2
        E10:    float   [MeV]  Beam total-energy for WS1
        Th1:    float   [um]   Target thick. for WS1
        <Ret> 
        Th2:    float   [um]   dE [MeV/A] equiv. Thick. for WS2
            = NaN   if (E10 <= 0) | (E11 <= 0) | (Th1 < 0)     
            = 0     if (Th1 = 0)
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        Th2 = self._Enew_eq_Th('dEu_eq_Th',WSnm1,WSnm2,E10,Th1)
        return( Th2 )

    def DEu_eq_Th(self, WSnm1:str, WSnm2:str,
                 E10:float, Th1:float ) -> float:
        """ == dEu_eq_Th() """
        return( self.dEu_eq_Th(WSnm1,WSnm2,E10,Th1) )

    def dEu_eq_ThGas(self, WSnm1:str, WSnm2:str,
            E10:float, Tmm1:float, 
            Pa1:float, degC1:float, Pa2:float, degC2:float ) -> float:
        """ dEu [MeV/A] equivalent Thickness : Gas
            ie.) get Tmm2 [mm] of 
                dE1 = (E10        - EnewGas(WS1,E10,Tmm1,Pa1,degC1)) 
             == dE2 = (E20 == E10 - EnewGas(WS2,E20,Tmm2,Pa2,degC2))
             ie.) this is identical with Enew_eq_ThGas()
        WSnm1:  str     srWS name for E10, Th1
        WSnm2:  str     srWS name for E20, Th2
        E10:    float   [MeV]  Beam total-energy for WS1
        Th1:    float   [um]   Target thick. for WS1
        <Ret> 
        Tmm2:   float   [mm]   dE [MeV/A] equiv. Thick. for WS2
            = NaN   if (E10 <= 0) | (E11 <= 0) | (Th1 < 0)     
            = 0     if (Th1 = 0)
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        msgH = 'dEu_eq_ThGas'
        kG1,kG2 = self._Gas_eq_Gas(msgH,WSnm1,WSnm2,Pa1,degC1,Pa2,degC2)             
        Th1 = Tmm1 * kG1
        Th2 = self._Enew_eq_Th(msgH,WSnm1,WSnm2,E10,Th1) 
        Tmm2= Th2 / kG2
        return( Tmm2 )

    def DEu_eq_ThGas(self, WSnm1:str, WSnm2:str,
            E10:float, Tmm1:float, 
            Pa1:float, degC1:float, Pa2:float, degC2:float ) -> float:
        """ == dEu_eq_ThGas() """
        return( self.dEu_eq_ThGas(WSnm1,WSnm2,E10,Tmm1,Pa1,degC1,Pa2,degC2) )

    """ Two srWS combination func. : dEt_eq_Th() dEt_eq_ThGas()
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    
    def dEt_eq_Th(self, WSnm1:str, WSnm2:str,
                 E10:float, Th1:float ) -> float:
        """ dEu * A:total_E [MeV] equivalent Thickness
            ie.) get Th2 [um] of 
                dE1 = (E10        - Enew(WS1,E10,Th1); E11 ) * A1 
             == dE2 = (E20 == E10 - Enew(WS2,E20,Th2); E21 ) * A2
        WSnm1:  str     srWS name for E10, Th1
        WSnm2:  str     srWS name for E20, Th2
        E10:    float   [MeV]  Beam total-energy for WS1
        Th1:    float   [um]   Target thick.     for WS1
        <Ret> 
        Th2:    float   [um]   dE [MeV] equiv. Thick. for WS2
            = NaN   if (E10 <= 0) | (E11 <= 0) | (Th1 < 0)        
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        Th2 = self._dEt_eq_Th('dEt_eq_Th',WSnm1,WSnm2,E10,Th1)
        return( Th2 )

    def DEt_eq_Th(self, WSnm1:str, WSnm2:str,
                 E10:float, Th1:float ) -> float:
        """ == dEt_eq_Th() """
        return( self.dEt_eq_Th(WSnm1,WSnm2,E10,Th1) )

    def dEt_eq_ThGas(self, WSnm1:str, WSnm2:str,
                E10:float, Tmm1:float, 
                Pa1:float, degC1:float, Pa2:float, degC2:float ) -> float:
        """ dEu * A:total_E [MeV] equivalent Thickness : Gas
            ie.) get Tmm2 [mm] of 
                dE1 = (E10        - EnewGas(WS1,E10,Tmm1,Pa1,degC1)) * A1 
             == dE2 = (E20 == E10 - EnewGas(WS2,E20,Tmm2,Pa2,degC2)) * A2
        WS1:    str     srWS name for E10, Th1
        WS2:    str     srWS name for E20, Th2
        E10:    float   [MeV]  Beam total-energy for WS1
        Tmm1:   float   [mm]    Target thickness for WS1
        Pa1:    float   [Pa]    Gas pressure     for WS1
        degC1:  float   [degC]  Gas temperature  for WS1
        Pa2:    float   [Pa]    Gas pressure     for WS2
        degC2:  float   [degC]  Gas temperature  for WS2
        <Ret> 
        Tmm2    float   [mm]   Enew equiv. Thick. for WS2
            = NaN   if (E10 <= 0) | (E11 <= 0) | (Th1 < 0)        
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        msgH = 'dEt_eq_ThGas'
        kG1,kG2 = self._Gas_eq_Gas(msgH,WSnm1,WSnm2,Pa1,degC1,Pa2,degC2)             
        Th1 = Tmm1 * kG1
        Th2 = self._dEt_eq_Th(msgH,WSnm1,WSnm2,E10,Th1) 
        Tmm2= Th2 / kG2
        return( Tmm2 )

    def DEt_eq_ThGas(self, WSnm1:str, WSnm2:str,
                E10:float, Tmm1:float, 
                Pa1:float, degC1:float, Pa2:float, degC2:float ) -> float:
        """ == dEt_eq_ThGas() """
        return( self.dEt_eq_ThGas(WSnm1,WSnm2,E10,Tmm1,Pa1,degC1,Pa2,degC2) )

    def _dEt_eq_Th(self, msgH:str, WSnm1:str, WSnm2:str,
            E10:float, Th1:float) -> float:
        """ dE * A:total_E [MeV] equivalent Thickness : common
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        ws1,ws2= self.getWS(WSnm1), self.getWS(WSnm2)
        A1,A2   = ws1.IonA, ws2.IonA
        if(E10 <= 0.):
            msgL= self._msgW_E10_lt0(msgH,E10,'dE1')
            _Err(1,'\n'.join(msgL))
            return( float('NaN') )
        if(Th1 < 0):                    # Th1 = 0 is valid
            msg = self._msgW_Th_lt0(msgH,'Th1',Th1)
            _Err(1,msg)
            return( float('NaN') )
        E11= ws1._Enew(E10,Th1)         # use _Enew() to avoid NOT Gas msg.
        if(E11 <= 0.):
            msgL= self._msgW_Enew1_lt0(msgH,WSnm1,E10,Th1,E11,'dE2')
            _Err(1,'\n'.join(msgL))
            return( float('NaN') )
        dE1t= (E10 - E11)*A1
        E20 = E10
        R20 = ws2.Rng(E20)  # Warn: R20 out of ws2._Rng ; checked by ws.Rng()
        dE2t= dE1t
        dE2 = dE2t / A2
        E21 = E20 - dE2
        if(E21 <= 0.):
            ms1= '#@ {}:Warn:'.format(msgH) 
            ms2= self._msgW_Enew(WSnm1,'E10','Th1',E10,Th1,E11)+' = E11'
            ms3=('   dE1t = (E10[{:.2f}] - E11[{:.2f}]) '
                 '* A1[{:3d}] = [{:.2f}] MeV').format(E10,E11,A1,dE1t)
            ms4=('   dE2u = dE2t / A2 = dE1t[{:.2f}] / A2[{:3d}] '
                 '= [{:.2f}] MeV/A').format(dE1t,A2,dE2)
            ms5=('   E21 = E20 - dEu2 = E10[{:.2f}] - dE2u[{:.2f}] < 0'
                 ).format(E10,dE2)
            ms6= ' ; then there is no answer.'
            _Err(1,'\n'.join([ms1,ms2,ms3,ms4,ms5,ms6]))
            return( float('NaN') )
        R21 = ws2.Rng(E21)  # Warn: R21 out of ws2._Rng ; ''
        Th2 = R20 - R21
        return( Th2 )

    """ Two srWS combination func. : dEu_eq_Eold() dEu_eq_EoldGas()
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""

    def dEu_eq_Eold(self, WSnm1:str, WSnm2:str,
                 E10:float, Th1:float, Hside:bool) -> float:
        """ dEu [MeV/A] equivalent Eold
            ie.) get E20 [MeV/A] of 
                dE1 = (E10 - Enew(WS1,E10,Th1)       ; E11 ) 
             == dE2 = (E20 - Enew(WS2,E20,Th2 == Th1); E21 )
        WSnm1:  str     srWS name for E10, Th1
        WSnm2:  str     srWS name for E20, Th2
        E10:    float   [MeV]  Beam total-energy for WS1
        Th1:    float   [um]   Target thick.     for WS1
        Hside:  bool    if(True) search at E-higher side of the LETe peak
                        else      ''       E-lower-side  ''
        <Ret> 
        E20:    float   [MeV/A]   dEu equiv. E   for WS2
            = NaN   if (E10 <= 0) | (Th1 <= 0) | dE2=dE1 not found
        cf.) E11=0: stop is OK 
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        Th2 = Th1
        E20 = self._dEu_eq_Eold('dEu_eq_Eold',WSnm1,WSnm2,E10,Th1,Th2,Hside)
        return( E20 )
        
    def DEu_eq_Eold(self, WSnm1:str, WSnm2:str,
                 E10:float, Th1:float, Hside:bool=True) -> float:
        """ == dEu_eq_Eold() """
        return( self.dEu_eq_Eold(WSnm1,WSnm2,E10,Th1,Hside) )

    def dEu_eq_EoldGas(self, WSnm1:str, WSnm2:str,
                E10:float, Tmm1:float, Hside:bool, 
                Pa1:float, degC1:float, Pa2:float, degC2:float ) -> float:
        """ dEu [MeV/A] equivalent Eold : Gas
            ie.) get E20 [MeV/A] of 
                dE1 = (E10 - EnewGas(WS1,E10,Tmm1,        Pa1,degC1); E11) 
             == dE2 = (E20 - EnewGas(WS2,E20,Tmm2 == Tmm1,Pa2,degC2); E21)
        WSnm1:  str     srWS name for E10, Th1
        WSnm2:  str     srWS name for E20, Th2
        E10:    float   [MeV]  Beam total-energy for WS1
        Tmm1:   float   [mm]    Target thickness for WS1
        Hside:  bool    if(True) search at E-higher side of the LETe peak
                        else      ''       E-lower-side  ''
        Pa1:    float   [Pa]    Gas pressure     for WS1
        degC1:  float   [degC]  Gas temperature  for WS1
        Pa2:    float   [Pa]    Gas pressure     for WS2
        degC2:  float   [degC]  Gas temperature  for WS2
        <Ret> 
        E20:    float   [um]   dE [MeV/A] equiv. E [MeV/A] for WS2
            = NaN   if (E10 <= 0) | (Th1 <= 0) | dE2=dE1 not found
        cf.) E11=0: stop is OK 
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        msgH = 'dEu_eq_EoldGas'
        kG1,kG2 = self._Gas_eq_Gas(msgH,WSnm1,WSnm2,Pa1,degC1,Pa2,degC2)             
        Th1 = Tmm1 * kG1
        Th2 = Tmm1 * kG2
        E20 = self._dEu_eq_Eold(msgH,WSnm1,WSnm2,E10,Th1,Th2,Hside) 
        return( E20 )

    def DEu_eq_EoldGas(self, WSnm1:str, WSnm2:str,
                E10:float, Tmm1:float, Hside:bool, 
                Pa1:float, degC1:float, Pa2:float, degC2:float ) -> float:
        """ == dEu_eq_EoldGas() """
        return( self.dEu_eq_EoldGas(WSnm1,WSnm2,E10,Tmm1,Hside, 
                        Pa1,degC1,Pa2,degC2) )

    def _dEu_eq_Eold(self, msgH:str, WSnm1:str, WSnm2:str,
            E10:float, Th1:float, Th2:float, Hside:bool=True) -> float:
        """ dE [MeV/A] equivalent Eold : common
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        ws1,ws2 = self.getWS(WSnm1), self.getWS(WSnm2)
        if(E10 <= 0.):
            msgL= self._msgW_E10_lt0(msgH,E10,'dE1')
            _Err(1,'\n'.join(msgL))
            return( float('NaN') )
        if(Th1 <= 0):                   # Th1 = 0 is NOT valid
            msg = self._msgW_Th_lt0(msgH,'Th1',Th1)
            _Err(1,msg)
            return( float('NaN') )
        E11= ws1._Enew(E10,Th1)         # use _Enew() to avoid NOT Gas msg.
        """ E11 <= 0 : STOP is OK """
        #if(E11 <= 0.):
        #    msgL= self._msgW_Enew1_lt0('dEu_eq_Eold',WSnm1,E10,Th1,E11,'dE2')
        #    _Err(1,'\n'.join(msgL))
        #    return( float('NaN') )
        dE1 = E10 - E11
        dE2 = dE1
        E20,dEmax= self._interp_E_dE(ws2, Th2, dE2, Hside, A=1)
        if(E20 is None):
            ms1= '#@ {}:Warn:'.format(msgH) 
            ms2= self._msgW_Enew(WSnm1,'E10','Th1',E10,Th1,E11)+' = E11'
            ms3= '   dE1 = (E10[{:.2f}] - E11[{:.2f}]) = [{:.2f}] MeV/A'.format(
                E10,E11,dE1)
            ms4= ' while dE2.max[{:.2f}] < dE1[{:.2f}] '.format(dEmax,dE1)
            ms5= ' ; then there is no answer.'
            _Err(1,'\n'.join([ms1,ms2,ms3,ms4,ms5]))
            return( float('NaN') )
        return( E20 )

    def _interp_E_dE(self, ws:object, Th:float,
                     dEg:float, Hside:bool, A:int) -> float:
        """ make E vs dE table & interp at dEg
        make    dE[] = ( ws._E[] - ws.Enew(ws._E[],Th) ) * A
        find    dEgoal in dE[]
        <Ret>   
        E       [MeV/A] or None if not found
        dEmax   [MeV/A] max of dE[]
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        E0 = copy.copy( ws._E )
        E1 = [ws._Enew(E,Th) for E in E0]   # use _Enew() to avoid NOT Gas msg.
        dE = (E0 - E1) * A
        dEmax= dE.max()
        if(dEmax < dEg):                # dE[] peak is less than dEg
            return( None, dEmax )           # No answer
        """  separate (E[] vs dE[]) at Lower/Higher-side of its dE-peak """
        ixPk  = dE.argmax()
        E0L,E0H = E0[0:ixPk+1], E0[ixPk:]
        dEL,dEH = dE[0:ixPk+1], dE[ixPk:]
        E0H,dEH = E0H[::-1],   dEH[::-1]    # inverse-sort for H-side
        if(False):                          # for Dubug 
            print('A,dEg,ixPk=',A,dEg,ixPk)
            for i,(_e0,_e1,_de) in enumerate(zip(E0,E1,dE)):
                print('{:3d}, {:7.2f}, {:7.2f}, {:7.2f}'.format(i,_e0,_e1,_de) )
        if(Hside):  return( np.interp(dEg, dEH, E0H), dEmax )
        else:       return( np.interp(dEg, dEL, E0L), dEmax )
        
    """ Two srWS combination func. : dEt_eq_Eold() dEt_eq_EoldGas()
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    
    def dEt_eq_Eold(self, WSnm1:str, WSnm2:str,
                 E10:float, Th1:float, Hside:bool=True) -> float:
        """ dEt [MeV] equivalent Eold
            ie.) get E20 [MeV/A] of 
                dE1t = (E10 - Enew(WS1,E10,Th1)       ; E11 ) * A1
             == dE2t = (E20 - Enew(WS2,E20,Th2 == Th1); E21 ) * A2
        WSnm1:  str     srWS name for E10, Th1
        WSnm2:  str     srWS name for E20, Th2
        E10:    float   [MeV]  Beam total-energy  for WS1
        Th1:    float   [um]   Target thick. for WS1
        <Ret> 
        E20:    float   [MeV/A] dEt equiv. E for WS2
            = NaN   if (E10 <= 0) | (Th1 <= 0) | dE2=dE1 not found
        cf.) E11=0: stop is OK 
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        Th2 = Th1        
        E20 = self._dEt_eq_Eold('dEt_eq_Eold',WSnm1,WSnm2,E10,Th1,Th2,Hside)
        return( E20 )

    def DEt_eq_Eold(self, WSnm1:str, WSnm2:str,
                 E10:float, Th1:float, Hside:bool=True) -> float:
        """ == dEt_eq_Eold() """
        return( self.dEt_eq_Eold(WSnm1,WSnm2,E10,Th1,Hside) )

    def dEt_eq_EoldGas(self, WSnm1:str, WSnm2:str,
                 E10:float, Tmm1:float, Hside:bool, 
                 Pa1:float, degC1:float, Pa2:float, degC2:float ) -> float:
        """ dEt [MeV] equivalent Eold : Gas
            ie.) get E20 [MeV/A] of 
                dE1t = (E10 - Enew(WS1,E10,Th1,      ,Pa1,degC1) ; E11 ) * A1
             == dE2t = (E20 - Enew(WS2,E20,Th2 == Th1,Pa2,degC2) ; E21 ) * A2
        WSnm1:  str     srWS name for E10, Tmm1
        WSnm2:  str     srWS name for E20, Tmm2
        E10:    float   [MeV]  Beam total-energy for WS1
        Tmm1:   float   [um]   Target thick.     for WS1
        Hside:  bool    if(True) search at E-higher side of the LETe peak
                        else      ''       E-lower-side  ''
        Pa1:    float   [Pa]    Gas pressure     for WS1
        degC1:  float   [degC]  Gas temperature  for WS1
        Pa2:    float   [Pa]    Gas pressure     for WS2
        degC2:  float   [degC]  Gas temperature  for WS2
        <Ret> 
        E20:    float   [MeV/A] dEt equiv. E for WS2
            = NaN   if (E10 <= 0) | (Th1 <= 0) | dE2=dE1 not found
        cf.) E11=0: stop is OK 
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        msgH = 'dEt_eq_EoldGas'
        kG1,kG2 = self._Gas_eq_Gas(msgH,WSnm1,WSnm2,Pa1,degC1,Pa2,degC2)             
        Th1 = Tmm1 * kG1
        Th2 = Tmm1 * kG2
        E20 = self._dEt_eq_Eold(msgH,WSnm1,WSnm2,E10,Th1,Th2,Hside) 
        return( E20 )

    def DEt_eq_EoldGas(self, WSnm1:str, WSnm2:str,
                 E10:float, Tmm1:float, Hside:bool, 
                 Pa1:float, degC1:float, Pa2:float, degC2:float ) -> float:
        """ == dEt_eq_EoldGas() """
        return( self.dEt_eq_EoldGas(WSnm1,WSnm2,E10,Tmm1,Hside, 
                        Pa1,degC1,Pa2,degC2) )

    def _dEt_eq_Eold(self, msgH:str, WSnm1:str, WSnm2:str,
            E10:float, Th1:float, Th2:float, Hside:bool=True) -> float:
        """ dEt [MeV] equivalent Eold : common
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        ws1,ws2 = self.getWS(WSnm1), self.getWS(WSnm2)
        A1,A2   = ws1.IonA, ws2.IonA
        if(E10 <= 0.):
            msgL= self._msgW_E10_lt0(msgH,E10,'dE1')
            _Err(1,'\n'.join(msgL))
            return( float('NaN') )
        if(Th1 <= 0):                   # Th1 = 0 is NOT valid
            msg = self._msgW_Th_lt0(msgH,'Th1',Th1)
            _Err(1,msg)
            return( float('NaN') )
        E11= ws1._Enew(E10,Th1)         # use _Enew() to avoid NOT Gas msg.
        """ E11 <= 0 : STOP is OK """
        dE1  = (E10 - E11)
        dE1t = dE1 * A1
        dE2t = dE1t
        E20,dEmax= self._interp_E_dE(ws2, Th2, dE2t, Hside, A=A2)
        if(E20 is None):
            ms1= '#@ {}:Warn:'.format(msgH) 
            ms2= self._msgW_Enew(WSnm1,'E10','Th1',E10,Th1,E11)+' = E11'
            ms3=('   dE1t= (E10[{:.2f}] - E11[{:.2f}]) '
                 '= [{:.2f}] MeV/A * A1[{:3d}] = [{:.2f}]').format(
                E10,E11,dE1,A1,dE1t)
            ms4= ' while dE2t.max[{:.2f}] < dE1t[{:.2f}] '.format(dEmax,dE1t)
            ms5= ' ; then there is no answer.'
            _Err(1,'\n'.join([ms1,ms2,ms3,ms4,ms5]))
            return( float('NaN') )
        return( E20 )

    """ SrWB : info. WSname pivot_table
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""

    def info_pvt(self) -> object:
        """ make pivot-table of Beam vs Target
        <Ret> 
        pvt:    pd.DataFrame     pivot-table of Beam vs Target
        <HowTo> pvt = srWB.info_pvt()
                print(pvt.to_string(index=False))
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""        
        SpL = self.getWSnmLDB(swSort=1)
        df  = pd.DataFrame(SpL)
        #bmL,tgL = df['Bm'].unique(), df['Tg'].unique()
        """ pivot_table 作成                     ->r#401) 
            Multi index ['Z','Bm','Sym','A'] でsort """
        pt0 = df.pivot_table(index=['Z','Bm','Sym','A'],
                             columns='Tg',aggfunc='count')
        """ NaN を 0 に一括変換                   ->r#402) """
        pt1 = pt0.fillna(0)
        """ count=0 を x  =1 を o に 一括変換      ->r#403) """
        fmap = lambda x: 'o' if x>0 else 'x'
        pt2 = pt1.applymap(fmap)
        """ Multi index を解除                   ->r#404) """
        pt3 = pt2.reset_index()
        pt3t = pt3.T                # 転置して columns も Mul.ix 解除
        pt31 = pt3t.reset_index()
        pt31.iat[1,1] = 'Bm'
        pt31.iat[2,1] = 'Sym'
        pt31.iat[3,1] = 'A'
        pt32 = pt31.drop(columns=['level_0'])   # 列#1 削除
        pt32.rename(columns={'Tg':0})
        pt33 = pt32.drop(pt32.index[[0,1]])     # 行#0,1 削除
        #pd.set_option('display.max_columns', 200) # columns=列
        #print(pt33.to_string(index=False))
        # TODO 未解決
        """ pt33.columns が　未解決
        Index(['Tg', 0, 1, .., 20], 'Tg' で変？
              dtype='object')  <-- object で変？
        """
        return( pt33 )

    def _info_pvt_TEST1(self):
        SpL = self.getWSnm_SpL(swSort=1)
        df  = pd.DataFrame(SpL)
        #bmL,tgL = df['Bm'].unique(), df['Tg'].unique()
        """ pivot_table 作成                     ->r#401) 
            Multi index ['Z','Bm','Sym','A'] でsort """
        pt0 = df.pivot_table(index=['Z','Bm','Sym','A'],
                             columns='Tg',aggfunc='count')
        """ NaN を 0 に一括変換                   ->r#402) """
        pt1 = pt0.fillna(0)
        """ count=0 を x  =1 を o に 一括変換      ->r#403) """
        fmap = lambda x: 'o' if x>0 else 'x'
        pt2 = pt1.applymap(fmap)
        """ Multi index を解除                   ->r#404) """
        pt3 = pt2.reset_index()
        return( pt0,pt1,pt2,pt3 )

    def _info_pvt_TEST2(self):
        SpL = self.getWSnm_SpL(swSort=1)
        df  = pd.DataFrame(SpL)
        #bmL,tgL = df['Bm'].unique(), df['Tg'].unique()
        """ pivot_table 作成                     ->r#401) 
            Multi index ['Z','Bm','Sym','A'] でsort """
        pt0 = df.pivot_table(index=['Z','Bm','Sym','A'],
                             columns='Tg',aggfunc='count')
        """ NaN を 0 に一括変換                   ->r#402) """
        pt1 = pt0.fillna(0)
        """ count=0 を x  =1 を o に 一括変換      ->r#403) """
        fmap = lambda x: 'o' if x>0 else 'x'
        pt2 = pt1.applymap(fmap)
        pt2t= pt2.T
        pt21= pt2t.reset_index()
        pt22= pt21.T
        pt23= pt22.reset_index()
        pt24 = pt23.drop(columns=['Z','Bm'])
        pt25 = pt24.drop(pt24.index[[0]])     # 行#0,1 削除
        return( pt25 )

    """ srWS : info.
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""

    def WSinfo(self, WSnm:str, modL:list=[0], *, 
               intrvE:int=1, show=True):
        """ Info. of a srimWS
        WSnm:   str     srWS name        
        <optional>  see) srimWS.info()
        modL:   [int,..]    dump mode list
            0: simple info. ; wsName,Beam,Target
            1: other  info. ; TrgDens, BraggC, SRIMver, Corded
            2: target compo.; TrgCmp table
            3: conv. factor ; LET unit conversion table
            4: table size   ; {min|max} of E, LETe,n,t, Rng, StLng, StLtr
            5: table contens; E, LET{e|n|t}, Rng, StLng, StLtr
            6: table size   ; LET{e|n|t}{L|H}
           -1: all of above 
        <optional: kwd args>
        intrvE: int         when mod=5, skip E interval 
        show:   bool        show dump result as print()
        <Ret>   
        nothing             if show=True
        txL:    [str,..]    else dump result as list of text    
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        ws = self.getWS(WSnm)
        txL = ws.info(modL=modL, intrvE=intrvE, show=show)
        if(not show): 
            return( txL )

    def WSplot_LET(self, WSnm:str, *,
                       LETe:bool=True, LETn:bool=True, LETt:bool=True,
                       rngX:tuple=(1e-5,1e3), rngY:tuple=(1e-2,1e2)) -> tuple:
        """ Plot ._LET{e|n|t} of a srimWS
        WSnm:           str     srWS name        
        <optional: kwd args> see) srimWS.plotLET()
        LET{e|n|t}:     bool    plot SW for LET{e|n|t}
        rngX,rngY:      tuple   plot range (min,max)
        <Ret> fg,ax             figure, axies
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        ws = self.getWS(WSnm)
        fg,ax = ws.plotLET(LETe=LETe, LETn=LETn, LETt=LETt,
                           rngX=rngX, rngY=rngY)
        return( fg, ax )

    def WSplot_LETcmp(self, WSnmL:list, *,
                    LETe:bool=False, LETn:bool=False, LETt:bool=True,
                    rngX:tuple=(1e-5,1e3), rngY=(1e-2,1e2) ) -> tuple:
        """ Plot LET{e|n|t} comp. with plural srWS  
        WSnmL:          [str,]  srWS name        
        <optional: kwd args> see) srimWSplot_LETcmp()
        LET{e|n|t}:     bool    plot SW for LET{e|n|t}
        rngX,rngY:      tuple   plot range (min,max)
        <Ret> fg,ax             figure, axies
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        wsL = self.getWSL(WSnmL)
        fg,ax = srimWSplot_LETcmp(wsL,LETe=LETe, LETn=LETn, LETt=LETt,
                           rngX=rngX, rngY=rngY)
        return( fg, ax )

    def WSplot_RngSt(self, WSnm:str, *,
                        Rng:bool=True, StLng:bool=True, StLtr:bool=True,
                        rngX:tuple=(1e-5,1e3), rngY=(1e-3,1e6) ) -> tuple:
        """ Plot {._Range|._StLng|._StLtr} of a srimWS
        WSnm:           str     srWS name        
        <optional: kwd args> see) srimWS.plotRngSt()
        Rng,StLng,StLtr:bool    plot SW for Rng,StLng,StLtr
        rngX,rngY:      tuple   plot range (min,max)
        <Ret> fg,ax             figure, axies
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        ws = self.getWS(WSnm)
        fg,ax = ws.plotRngSt(Rng=Rng, StLng=StLng, StLtr=StLtr,
                           rngX=rngX, rngY=rngY )
        return( fg, ax )

    def WSplot_RngStcmp(self, WSnmL:list, *,
                        Rng:bool=True, StLng:bool=False, StLtr:bool=False,
                        rngX:tuple=(1e-5,1e3), rngY=(1e-3,1e6) ) -> tuple:
        """ Plot Range, StLng, StLtr comp. with plural srWS
        WSnmL:          [str,]  srWS name        
        <optional: kwd args>
        Rng,StLng,StLtr:bool    plot SW for Rng,StLng,StLtr
        rngX,rngY:      tuple   plot range (min,max)
        <Ret> fg,ax             figure, axies
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        wsL = self.getWSL(WSnmL)
        fg,ax = srimWSplot_RngStcmp(wsL,Rng=Rng, StLng=StLng, StLtr=StLtr,
                           rngX=rngX, rngY=rngY)
        return( fg, ax )

    """ srWB : info. / dump
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""

    def info(self, mod:int=1, *, 
             show:bool=True ) -> list:
        """ info. of this srWB
        mod:    int         dump mode
            0: simple info. ; .WBpath, .nWS, .curWSname
            1: ++ pivot-table of Beam vs Target
        <optional: kwd args>
        show:   bool        show dump result as print()        
        <Ret>   
        nothing             if show=True
        txL:    [str,..]    else dump result as list of text    
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        tx1 = 'srimWB.WBpath= [{}]'.format(self.WBpath)
        tx2 = '      .nWS= [{:3d}] sheets included.'.format(self.nWS)
        tx3 = '      .curWSname= [{}]'.format(self.curWSname)
        txL = [tx1,tx2,tx3]
        if(mod > 0):
            txL+= self.WBinfo(show=False)
        if(show):   print('\n'.join(txL))
        return( txL )

    def WBinfo(self, *, 
               show=True) -> list:
        """ srWB info. : show pivot-table of Beam vs Target
        <optional: kwd args>
        show:   bool        show dump result as print()        
        <Ret>   
        nothing             if show=True
        txL:    [str,..]    else dump result as list of text    
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        pvt = self.info_pvt()
        txt = pvt.to_string(index=False)
        txL = txt.splitlines()      # txt with '\n' -> list
        if(show):   print(txt)
        return(txL)

    def WBdump(self, wsnmL:list=None, *,
               fnOut:str='_srimWBdump.ppt',
               show:bool=True, closefg:bool=True ):
        """ srimWB contents dump to Power-Point file 
        <optional>
        wsnmL:  [str]       restrict srimWS to dump. if None, use srWB.WSnameL
        <optional: kwd args>
        fnOut:  str         output .ppt file name
        show:   bool        show　progress message
        closefg:bool = True, if plot > 20 figures; no plot in plot-window
                     = False,for check plot-window
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        srimWBdump(self, wsnmL=wsnmL,fnOut=fnOut,show=show,closefg=True)
     
    """ srWB : for print()
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    def __str__(self):
        txL = self.info( mod=0, show=False )
        return('\n'.join(txL))

 
""" ========================================================
            Test Main
    ========================================================"""
if __name__ == '__main__':
    from  pprint  import  pprint   # ->r#301  

    """ Test-CTRL Sw ---.---.---.---.---.---.---.---.---"""
    SwTest01 = True     # very simple examples
    SwTest11 = False    # srWB info

    """ SRIM DB
    注） DB_SRIM/MySRIMwb.pickle は、最新ver.の pickle にしておく事
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    DB_SRIM = './_MySRIMwb'
    
    # load MySRIMwb.pickle
    srWB= srimWB()
    srWB.loadWS(path=DB_SRIM)
    print(srWB)
    
    """ test01 : HowToUse ; very simple examples
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    if(SwTest01):
        WSnm = 'srim40Ar_Si'
        WSnm = 'srim12C_Water'
        srWS = srWB.getWS(WSnm)
        print('{} maxLETt= {} [{}]'.format(
            WSnm, srWB.maxLETt(WSnm), srWB.LETunit(0) ) )   # srWB usage
        srWS = srWB.getWS(WSnm)
        print('{} maxLETt2E= {}[MeV/A]'.format(
            WSnm, srWS.maxLETt2E ) )                        # seWS usage
        E = srWS.LETt2E(122,Uid=2,Hside=True)
        
        
    """ test11 : srWB info.
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    if(SwTest11):
        wsnmL = srWB.getWSnmL(Beam='Kr',Trg='Si',swSort=1)
        print(wsnmL)
        wsnmLDB = srWB.getWSnmLDB(Beam='Kr',swSort=1)
        pprint(wsnmLDB,width=100,sort_dicts=False)

    print('@@main: -- Fin ---')


""" ========================================================
        ref) Numbered ref
    ========================================================    
// pickle //
r#101) 
    Qiita2016 PythonのPickleモジュールを理解したいだけの人生だった
〇  https://qiita.com/hatt0519/items/f1f4c059c28cb1575a93
    Python: オブジェクトを漬物 (Pickle) にする
〇  https://blog.amedama.jp/entry/2015/12/05/132520

// sort //
r#201) nkmk Pythonで辞書のリストを特定のキーの値に従ってソート
    https://note.nkmk.me/python-dict-list-sort/
    引数keyにoperator.itemgetter()を指定
    複数のキーを基準にソート

// pprint //
r#301) 
〇  nkmk Pythonのpprintの使い方（リストや辞書を整形して出力）
    https://note.nkmk.me/python-pprint-pretty-print/
〇  8.11. pprint — データ出力の整然化 — Python 3.6.5 ドキュメント
    https://note.nkmk.me/python-pprint-pretty-print/

// pandas //
r#401) DeepAge Pandasでピボットテーブルを手軽に作成する
    https://deepage.net/features/pandas-pivot.html
〇  Python Data Science Handbook 3.10　ピボットテーブル
〇  https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.pivot_table.html
〇 　キノブログ Pandas入門講座｜10.データ集計（pivot_table）の方法
    https://kino-code.com/pandas_pivot_table/

r#402) nkmk pandasで欠損値NaNを置換（穴埋め）するfillna
    https://note.nkmk.me/python-pandas-nan-fillna/

r#403) nkmk pandasで要素、行、列に関数を適用するmap, applymap, apply
    https://note.nkmk.me/python-pandas-map-applymap-apply/

r#404) nkmk pandasのMultiindexの指定・追加・解除・ソート・レベル変更
    https://note.nkmk.me/python-pandas-multiindex-set-reset-sort-swap/

r#405) Qiita Pandasユーザーガイド「マルチインデックス・高度な索引」（公式ドキュメント日本語訳）
    https://qiita.com/nkay/items/63afdd4e96f21efbf62b
〇 　Yutaka Multiindex図解で解説｜基本構造・変更・追加・解除・並べ替え
    https://www.yutaka-note.com/entry/pandas_multiind_set
    

    pandas groupbyでグループ化
    https://www.yutaka-note.com/entry/pandas_groupby

    Qiita Pandasユーザーガイド「Group by - 分割・適用・結合」（公式ドキュメント日本語訳）
    https://qiita.com/nkay/items/322eebbe3bacfc451269

    ========================================================
        ref) Others
    ======================================================== 
    
// メモリ管理 //
r#901) 静かなる名辞 pythonでメモリ不足になったときにすること
    https://www.haya-programming.com/entry/2017/02/09/190512

    
------------------------------------------------------------------
"""