# -*- coding: utf-8 -*-
"""
    Module: pySRIMfit
    File:   srimWBcnv.py
    Descr.: srimWB converter helper funcs.
    Update: 220403 build start Ay
            220523 v.1.0 done 
    Last:   230318 v.1.1 refactored
    
@author: Ayoshida(RIKEN,Nishina center)
"""

""" import: standard lib.       
----.---.---.---.---.---.---.---.---.---.---.---.---"""
import  os
import  sys
import  xlwings as  xw

""" import: my custom lib.       
----.---.---.---.---.---.---.---.---.---.---.---.---"""
#import  aypylib             as ay 

""" import: from this package   
----.---.---.---.---.---.---.---.---.---.---.---.---"""
from    pySRIMfit   import  (srimWB,srimWScnv)

""" .===.===.===.===.===.===.===.===.===.===.===.===.===.===.===.===
        Work area for Helper funcs
====.===.===.===.===.===.===.===.===.===.===.===.===.===.===.=== === """

""" module Constants
    MySRIMwb.xlsx を生成する為の雛形　./_proto に置いてある
----.---.---.---.---.---.---.---.---.---.---.---.---"""
#__ Input_prototype: for Excel SRIMwb
_fn_SRIMwb_Pty = './_proto/_MySRIMwb_prtyp.xlsx' # file name
_sn_SRIMwb_Pty = '_prtyp_'                       # sheet name

def _SRIMwb_proto( xlfnOut:str ) -> object:
    """ copy SRIMwb proto-type and save it as xlfnOut
    xlfnOut: str    output file name
    <Ret> (xwB,shnm)
    xwB  xw.Book()  proto-type book; contains no data sheets
    shnm    str     proto-type sheet name
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""        
    #print('@@srimWScnv_py2xl:dirname=',os.path.dirname(__file__))
    fnPty = os.path.join(os.path.dirname(__file__),_fn_SRIMwb_Pty)
    xwB = xw.Book(fnPty)            # open prototype WB
    xwB.save(xlfnOut)               # save it as Output file name
    xwB.close()
    xwB = xw.Book(xlfnOut)          # re-open it
    return( xwB, _sn_SRIMwb_Pty )

""" module Globals
----.---.---.---.---.---.---.---.---.---.---.---.---"""
xlapp = None

def _xlapp_hide():
    """ set xlwings be silent 
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""        
    global xlapp
    xlapp = xw.App(visible=False)   # コレを無しにすると、読込状況が見えます
    xlapp.display_alarts = False

def _xlapp_kill():
    """ kill xlwings process 
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""        
    global xlapp
    xlapp.quit()
    xlapp.kill()
  
""" .===.===.===.===.===.===.===.===.===.===.===.===.===.===.===.===
        Helper funcs for srimWScnv
====.===.===.===.===.===.===.===.===.===.===.===.===.===.===.=== === """
def srimWBcnv_xl2pick(xlfnL:list, dirOut:str='./', *, 
                      fnOut:str=None, swSave=True) ->object:
    """ Excel WkBook -> .pickle conversion
    xlfnL: [fn_path] input:  path list of Excel work books
    <optional>
    dirOut: str      output: path for 'MySRIMwb.pickle' None=curr.dir
    <optional: kwd args>
    fnOut:  str      output: file name defa='MySRIMwb.pickle'
    swSave: bool     if False, no save .pickle
    <Ret>   srimWB   converted srimWB
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""        
    msgH = 'srimWBcnv_xl2pick'
    _xlapp_hide()
    srWB = srimWB()                      # make a srimWB instance
    for fnIn in xlfnL:
        print('@@ {}: converting ...'.format(msgH),fnIn)
        xwB = xw.Book(fnIn)                 # open MySRIMwb
        for i,xws in enumerate(xwB.sheets):
            if(srWB.xlWSid not in xws.name): continue
            swsc = srimWScnv()              # make a srimWScnv instance
            sws  = swsc.xlWS2srWS(xws)      # convert xws to srimWS
            print(i,')',sws)
            #sws.info()
            srWB._appendWS( sws )           # append srimWS to srimWB
    if(swSave):
        print('@@ saving {} ...'.format(fnOut))
        srWB.saveWS(path=dirOut,fnm=fnOut)  # save as pickle
    xwB.close()
    print('@@ {}: done'.format(msgH))
    _xlapp_kill()
    return( srWB )

def srimWBcnv_xl2srWB(xlfnL:list) ->object:
    """ Excel WkBook -> srWB
    xlfnL: [fn_path] input:  path list of Excel work books
    <Ret>   srimWB   converted srimWB
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""        
    msgH = 'srimWBcnv_xl2srWB'
    print('@@ {}: converting ...'.format(msgH))
    srWB = srimWBcnv_xl2pick(xlfnL, swSave=False)
    print('@@ {}: done'.format(msgH))
    return( srWB )

""" srimWBcnv_srWB2pick() does NOT need,
        use srWB.saveWS(path,fnm)           """

def srimWBcnv_srWB2xl(srWB, xlfnOut:str='./MySRIMwb_out.xlsx'):
    """ srimWB --> Excel WkBook conversion
    srWB:    srimWB input:  srimWB in memory
    <optional>
    xlfnOut: str    output: file name; defa. './MySRIMwb_out.xlsx'
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""        
    msgH = 'srimWBcnv_srWB2xl'
    print('@@ {}: converting ...'.format(msgH))
    _xlapp_hide()
    xwB,snPty = _SRIMwb_proto( xlfnOut )    # open&save proto-type WB
    swsc = srimWScnv()                      # make srimWScnv instance
    for i,sws in enumerate(srWB.WSL):
        print(i,')',sws)
        snm = sws.wsName
        xwS= xwB.sheets[snPty].copy(name=snm)   # copy prtyp sheet to the end
        swsc.srWS2xlWS(sws,xwS)                 # convert sws to Excel WS
    xwB.sheets[snPty].delete()              # remove prtyp sheet
    print('@@ saving {} ...'.format(xlfnOut))
    xwB.save(xlfnOut)                       # save converted result
    xwB.close()
    print('@@ {}: done'.format(msgH))
    _xlapp_kill()

def srimWBcnv_pick2xl(dirIn:str=None, xlfnOut:str='./MySRIMwb_out.xlsx'):
    """ pySRIM WkBook --> Excel WkBook conversion
    <optional>
    dirIn:   str input:  path for 'MySRIMwb.pickle' None=srimWB().WBpath
    xlfnOut: str output: file name defa. './MySRIMwb_out.xlsx'
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""        
    msgH = 'srimWBcnv_pick2xl'
    srWB = srimWB()                  # make a srimWB instance
    if(not srWB.loadWS(path=dirIn)): # load SRIMwb.pickle
        print('@@ {}:Error: MySRIMwb.pickle invalid'.format(msgH))
        sys.exit()
    print('@@ {}: converting ...'.format(msgH))
    srimWBcnv_srWB2xl(srWB,xlfnOut=xlfnOut)
    print('@@ {}: done'.format(msgH))

def srimWBcnv_out2xl(outfnL:list, xlfnOut:str='./MySRIMwb_out.xlsx'):
    """ SRIM-2013.out(*.txt) -> Excel WkBook conversion
    outfnL: [str,]  input:  path list of SRIM.out files
    <optional>
    xlfnOut: str    output: file name defa. './MySRIMwb_out.xlsx'
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    msgH = 'srimWBcnv_out2xl'
    _xlapp_hide()
    xwB,snPty = _SRIMwb_proto( xlfnOut )    # open&save proto-type WB
    swsc = srimWScnv()                      # make srimWScnv instance
    for fnIn in outfnL:
        print('@@ {}: converting ...'.format(msgH),fnIn)
        if(not os.path.exists(fnIn)):   # check fnIn
            print('@@ err:{} not found'.format(fnIn))
            _xlapp_kill()
            return
        xwS= xwB.sheets[snPty].copy(name='new')   # copy prtyp sheet to the end
        swsc.srOut2xlWS(fnIn,xwS)                 # convert sws to Excel WS
    xwB.sheets[snPty].delete()      # remove prtyp sheet
    print('@@ saving {} ...'.format(xlfnOut))
    xwB.save(xlfnOut)               # save converted result
    xwB.close()
    print('@@ {}: done'.format(msgH))
    _xlapp_kill()

def srimWBcnv_csv2xl(csvfnL:list, xlfnOut:str='./MySRIMwb_out.xlsx'):
    """ SRIM-2013.out(*.xlsx) CSV-read WkBook -> Excel WkBook conversion
    csvfnL: [str,]  input:  path list of SRIM.out CSV work books
    <optional>
    xlfnOut: str    output: file name defa. './MySRIMwb_out.xlsx'
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    msgH = 'srimWBcnv_csv2xl'
    _xlapp_hide()
    xwB,snPty = _SRIMwb_proto( xlfnOut )    # open&save proto-type WB
    swsc = srimWScnv()                      # make srimWScnv instance
    for fnIn in csvfnL:
        print('@@ {}: converting ...'.format(msgH),fnIn)
        if(not os.path.exists(fnIn)):       # check fnIn
            print('@@ err:{} not found'.format(fnIn))
            _xlapp_kill()
            return
        xwBc = xw.Book(fnIn)
        for i,xwSc in enumerate(xwBc.sheets):
            print(i,')',xwSc.name)
            xwS = xwB.sheets[snPty].copy(name='new')  # copy prtyp sheet to the end
            swsc.srCSV2xlWS(xwSc,xwS)                 # convert sws to Excel WS
        xwBc.close()
    xwB.sheets[snPty].delete()      # remove prtyp sheet
    print('@@ saving {} ...'.format(xlfnOut))
    xwB.save(xlfnOut)               # save converted result
    xwB.close()
    print('@@ {}: done'.format(msgH))
    _xlapp_kill()


""" ========================================================
            Test Main
    ========================================================"""
if __name__ == '__main__':
    print('PLZ. use _test/t_srimWBcnv_*.__main__')
    print('@@main: -- Fin ---')
