# -*- coding: utf-8 -*-
"""
    Module: pySRIMfit
    File:   srimWBmisc.py
    Descr.: Miscellaneous class, func for srimWB
    Update: 220403 build start Ay
            220515 v.1.0 done
    Last:   230410 v.1.1 refactored
    
@author: Ayoshida(RIKEN,Nishina center)
"""

""" import: standard lib.       
----.---.---.---.---.---.---.---.---.---.---.---.---"""
import  matplotlib.pyplot   as plt  # used at plt.close(fg)

""" import: my custom lib.       
----.---.---.---.---.---.---.---.---.---.---.---.---"""
from    aypylibSbs      import  (pptRep)

""" import: from this package   
----.---.---.---.---.---.---.---.---.---.---.---.---"""
#import  e5ana               as e5 

""" .===.===.===.===.===.===.===.===.===.===.===.===.===.===.===.===
    srimWB : misc. func. def.
====.===.===.===.===.===.===.===.===.===.===.===.===.===.===.=== === """

""" .===.===.===.===.===.===.===.===.===.===.===.===.===.===.===.===
        Class   e5_PptRep       e5_*.py ppt report
====.===.===.===.===.===.===.===.===.===.===.===.===.===.===.=== === """

""" srimWB dump
----.---.---.---.---.---.---.---.---.---.---.---.---"""

class srWB_pptRep( pptRep ):
    """ Python-pptx helper class : simple ppt report
----.---.---.---.---.---.---.---.---.---.---.---.---.---"""

    """ class Attributes
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    def _clear_my_pptRep(self): 
        self.fnPpt  = str()     # output .ppt file name
        self.hdr1   = str()     # Header-1
        self.hdr2   = None      # Header-2
        self.ftr1   = None      # Footer-1
        self.ftr2   = None      # Footer-2
        self.pgN    = 0
        
        self._sld     = None    # current Slide()
        self._p_hdr1  = list()  # Header-1
        self._p_hdr2  = list()  # Header-2
        self._p_ftr1  = list()  # Footer-1 
        self._p_ftr2  = list()  # Footer-2 
        self._p_pgN   = list()  # Page No
                 
    def __init__(self, 
                 hdr1:str=None, hdr2:str=None):
        """  Constructor
        <optional>
        hdr1:   str     Header-1     eg) title
        hdr2:   str     Header-2     eg) srinWS.wsName
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""        
        super().__init__('A4','land')
        self._clear_my_pptRep()
        self.hdr1  = hdr1       # 
        self.hdr2  = hdr2
        self.set_Titles()
        self.fnPpt = '_srimWBdump.ppt'
    
    def set_Titles(self,prm:list=None):
        """  set page header/footer attributes
        prm:    list    [[(left,top),fs,fc,it,bd,ul],.. len=8]
                        if None, set default values
            where,
            (left,top)  [mm] Left,Top corner pos
            fs: int     [Pt] font size, if None use self.fntSiz
            fc: str          font color eg) 'w','r', ...
            it: bool         Italic
            bd: bool         Bold
            ul: bool         Under Line
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        if(prm is None):
            #               (left,top), fs,  fc, it,    bd,    ul
            self._p_hdr1  = [( 10,  5), 16, 'k', True,  False, False]
            self._p_hdr2  = [(150,  5), 16, 'k', True,  False, False]
            self._p_pgN   = [(250,200), 12, 'k', False, False, False]
        else:
            self._p_hdr1  = prm[3]
            self._p_hdr2  = prm[4]
            self._p_pgN   = prm[7]

    def newPage(self):
        """  add new page, write page titles
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        self._sld = self.addSlide()
        self.pgN += 1
        self.wr_Titles()
       
    def wr_Titles(self):
        """  write page header/fotter
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        def _wr_ttl(txt,txp):
            if(txt is not None):
                self.wrText(self._sld, txt, 
                            lt=txp[0], fs=txp[1], fc=txp[2],
                            it=txp[3], bd=txp[4], ul=txp[5])
        
        _wr_ttl(self.hdr1,  self._p_hdr1)
        _wr_ttl(self.hdr2,  self._p_hdr2)
        _wr_ttl('- '+str(self.pgN)+' -',  self._p_pgN)

    def wr_Text(self, txt:str, lt:tuple, *,
               fs=None, fc=None, fn=None,
               it:bool=False, bd:bool=False, ul:bool=False):
        """  write text box
        txt:    str          text to write
        lt:  (left,top) [mm] Left,Top corner pos
        <optional: kwd args>
        fs:     int     [Pt] font size, if None use self.fntSiz
        fc:     str          font color eg) 'w','r', ...
        fn:     str          font name, if None use self.fntNm
        it:     bool         Italic
        bd:     bool         Bold
        ul:     bool         Under Line
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        self.wrText(self._sld, txt, lt, fs=fs, fc=fc, fn=fn,
                    it=it,bd=bd,ul=ul)

    def wr_Image(self, im:object, lt:tuple, *,
                wid:int=None, hgt:int=None):
        """ write PIL Image
        im:     PIL.Image 
        lt:  (left,top) [mm] Left,Top corner pos
        <optional: kwd args>
        wid:    int     [mm] width
        hgt:    int     [mm] height
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        self.wrImage(self._sld, im, lt, wid=wid, hgt=hgt)

    def wr_Figure(self, fig:object, lt:tuple, *,
                wid:int=None, hgt:int=None):
        """ write matplot.figure
        fig:    matplotlib.Figure 
        lt:  (left,top) [mm] Left,Top corner pos
        <optional: kwd args>
        wid:    int     [mm] width
        hgt:    int     [mm] height
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        self.wrFigure(self._sld, fig, lt, wid=wid, hgt=hgt)
    
    def save_rep(self, fn:str=None):
        """ save report as .ppt
        fn:     str     save file name; defa) self.fnPpt
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        fn = self.fnPpt if(fn is None) else fn
        self.save_ppt( fn )


def srimWBdump( srWB:object, *, 
               wsnmL:list=None, fnOut:str='_srimWBdump.ppt', 
               show:bool=True, closefg:bool=True ):
    """ srimWB contents dump to .ppt 
    srWB:   srimWB      for dump
    <optional: kwd args>
    wsnmL:  [str]       restrict srimWS to dump. if None, use srWB.WSnameL
    fnOut:  str         output .ppt file name
    show:   bool        show　progress message
    closefg:bool = True, if plot > 20 figures; no plot in plot-window
                 = False,for check plot-window
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    rep = srWB_pptRep()
    # pg.1 : srimWB.info
    rep.hdr1 = 'srimWB.info'
    rep.newPage()
    txL = srWB.WBinfo(show=False)
    txt = '\n'.join(txL)
    rep.wr_Text( txt, ( 10, 12), fs=10, fc='b',bd=True )
    # select srWS to dump
    _wsnmL = srWB.WSnameL if(wsnmL is None) else wsnmL
    _wsL = srWB.getWSL(_wsnmL)
    for i,ws in enumerate(_wsL):
        rep.hdr1 = ws.wsName
        rep.newPage()
        if(show):
            print(' pg.{:3d} {}'.format(i,ws.wsName))
        # Beam, Target info
        txL = ws.info(modL=[0,1],show=False)
        rep.wr_Text( '\n'.join(txL), ( 10, 12), fs=10, fc='r',bd=True )
        # table size   ; {min|max} of E, LETe,n,t, Rng, StLng, StLtr
        txL = ws.info(modL=[4],show=False)
        rep.wr_Text( '\n'.join(txL), (  5, 35), fs= 8, fc='b',bd=True )
        # table contens; E, LET{e|n|t}, Rng, StLng, StLtr
        txL = ws.info(modL=[5],intrvE= 5,show=False)
        rep.wr_Text( '\n'.join(txL), (  5, 60), fs= 7, fc='k' )
        # TrgCmp table
        txL = ws.info(modL=[2],show=False)
        rep.wr_Text( '\n'.join(txL), (120, 12), fs= 9, fc='r',bd=True )
        # LET unit conversion table
        txL = ws.info(modL=[3],show=False)
        rep.wr_Text( '\n'.join(txL), (120, 60), fs= 8, fc='k',bd=True )
        # table size   ; LET{e|n|t}{L|H}
        #txL = ws.info(modL=[6],show=False)
        #rep.wr_Text( '\n'.join(txL), (120,100), fs=6, fc='k' )
        # plot LET{e|n|t}
        fg1,ax1 = ws.plotLET()
        rep.wr_Figure( fg1, lt=(190,  5), hgt=95 )
        # plot Range, StLng, StLtr
        fg2,ax2 = ws.plotRngSt()
        rep.wr_Figure( fg2, lt=(190,100), hgt=95 )
        if(closefg):
            plt.close(fg1)
            plt.close(fg2)
    if(show):
        print(' save to {}'.format(fnOut))
    rep.save_rep(fn=fnOut)
    
""" ========================================================
            Test Main
    ========================================================"""
if __name__ == '__main__':    
    print('@@main: -- Fin ---')


""" ========================================================
        ref) Numbered ref
    ========================================================
    
------------------------------------------------------------------
"""