# -*- coding: utf-8 -*-
"""
    Module: pySRIMfit
    File:   srimWSmisc.py
    Descr.: Miscellaneous class, func for srimWS
    Update: 220403 build start Ay
            220515 v.1.0 done
    Last:   230410 v.1.1 refactored
    
@author: Ayoshida(RIKEN,Nishina center)
"""

""" import: standard lib.       
----.---.---.---.---.---.---.---.---.---.---.---.---"""
import  re
import  matplotlib.pyplot   as plt
plt.style.use('default')
#plt.style.use('seaborn')
#plt.style.use('seaborn-whitegrid')
#plt.style.use('bmh')
#import  matplotlib.ticker   as  ticker
#import  matplotlib.cm       as cm
#from    matplotlib.colors   import  LogNorm
#from    matplotlib.gridspec import  GridSpec

""" import: my custom lib.       
----.---.---.---.---.---.---.---.---.---.---.---.---"""
#import  aypylib             as ay 

""" import: from this package   
----.---.---.---.---.---.---.---.---.---.---.---.---"""
#import  e5ana               as e5 

""" .===.===.===.===.===.===.===.===.===.===.===.===.===.===.===.===
    Debug CTRL
====.===.===.===.===.===.===.===.===.===.===.===.===.===.===.=== === """
""" Debug vs Error Level switch 
_DBG_Lev    msg         raise
    =0 ;    not-show    pass
    =1 ;    show        pass
    =2 ;    show        raise&stop
----.---.---.---.---.---.---.---.---.---.---.---.---"""
_DBG_Lev = 1        # current debug level

def _Err(errLev:int, msg:str):
    """ Error handle in this module
    errLev: int     Error level ; comp. with _DBG_Lev 
    msg:    str     raise message 
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    if(_DBG_Lev == 0): 
        return
    if(_DBG_Lev == 1):
        print(msg)
        return
    if((_DBG_Lev == 2) or (errLev > _DBG_Lev)):
        raise ValueError(msg)
    print(msg)

""" .===.===.===.===.===.===.===.===.===.===.===.===.===.===.===.===
        Atomic symbol
====.===.===.===.===.===.===.===.===.===.===.===.===.===.===.=== === """
""" import aypylib.atom_data.py """
class atomSym( object ):
    AT_AtomicSymbols = (        # def.as tuple
        'n', 'H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na', 'Mg',
        'Al', 'Si', 'P', 'S', 'Cl', 'Ar', 'K', 'Ca', 'Sc', 'Ti', 'V', 'Cr',
        'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn', 'Ga', 'Ge', 'As', 'Se', 'Br',
        'Kr', 'Rb', 'Sr', 'Y', 'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd',
        'Ag', 'Cd', 'In', 'Sn', 'Sb', 'Te', 'I', 'Xe', 'Cs', 'Ba', 'La',
        'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho', 'Er',
        'Tm', 'Yb', 'Lu', 'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir', 'Pt', 'Au',
        'Hg', 'Tl', 'Pb', 'Bi', 'Po', 'At', 'Rn', 'Fr', 'Ra', 'Ac', 'Th',
        'Pa', 'U', 'Np', 'Pu', 'Am', 'Cm', 'Bk', 'Cf', 'Es', 'Fm', 'Md',
        'No', 'Lr', 'Rf', 'Db', 'Sg', 'Bh', 'Hs', 'Mt', 'Ds', 'Rg', 'Cn',
        'Nh', 'Fl', 'Mc', 'Lv', 'Ts', 'Og'
        )
    def sym(self, Z:int):
        """ Z -> sym;       like 36 -> 'Kr'
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        if((Z < 0) or (Z > len(self.AT_AtomicSymbols))):
            raise ValueError('sym(Z) Z=',Z)
        return(self.AT_AtomicSymbols[Z])
    
    def Asym(self, A:int, Z:int ):
        """ A,Z -> Asym     like 36,84 -> '84Kr'
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        if((Z < 0) or (Z > len(self.AT_AtomicSymbols)) or
           (A < 0) or (A > 238)):
            raise ValueError('Asym(A,Z) A=',A,' Z=',Z)
        tx = '{0:3>}'.format(str(A))
        return(tx+self.AT_AtomicSymbols[Z])

    def Z(self, sym:str):
        """ sym -> Z        like 'Kr' -> 36
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        if(sym not in self.AT_AtomicSymbols):
            raise ValueError('Z(sym) sym=',sym)
        return(self.AT_AtomicSymbols.index(sym))


""" .===.===.===.===.===.===.===.===.===.===.===.===.===.===.===.===
    srimWS : Const. def.
====.===.===.===.===.===.===.===.===.===.===.===.===.===.===.=== === """
srimWS_TrgCmp = \
    ['AtomSym','AtomNum','Atom%','Mass%']

srimWS_Uid = \
    { 0:'MeV/(mg/cm2)',  1:'eV/Angstrom',    2:'keV/micron',
      3:'MeV/mm',        4:'keV/(ug/cm2)',   5:'MeV/(mg/cm2)',
      6:'keV/(mg/cm2)',  7:'eV/(1E15 atoms/cm2)', 8:'L.S.S. reduced unit'}

""" .===.===.===.===.===.===.===.===.===.===.===.===.===.===.===.===
    srimWS : misc. func. def.
====.===.===.===.===.===.===.===.===.===.===.===.===.===.===.=== === """

""" srimWS name
----.---.---.---.---.---.---.---.---.---.---.---.---"""

def srimWSname_AZT(ionA:int,ionZ:int,TgNS:str) -> str:
    """ make srimWS name by ionA, ionZ, TgNS
    ionA:   int     Beam A number;      like 84
    ionZ:   int     Beam Z number;      like 36
    TgNS:   str     Target short name;  like 'Air'
    <Ret>   str     WSname;             like 'srim84Kr_Air'
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    nm = 'srim'+atomSym().Asym(ionA,ionZ)+'_'+TgNS
    return( nm )

def srimWSname_ASymT(ionA:int,ionSym:str,TgNS:str) -> str:
    """ make srimWS name by ionA, ionSym, TgNS
    ionA:   int     Beam A number;      like 84
    ionSym: str     Beam Z symbol;      like 'Kr'
    TgNS:   str     Target short name;  like 'Air'
    <Ret>   str     WSname;             like 'srim84Kr_Air'
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    nm = 'srim'+str(ionA)+ionSym+'_'+TgNS
    return( nm )

def srimWSname_BmT(Beam:str,TgNS:str) -> str:
    """ make srimWS name by ionA, ionSym, TgNS
    Beam:   str     Beam name;          like '84Kr'
    TgNS:   str     Target short name;  like 'Air'
    <Ret>   str     WSname;             like 'srim84Kr_Air'
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    nm = 'srim'+Beam+'_'+TgNS
    return( nm )

def srimWSname_split(WSnm:str) -> list:
    """ split srimWS name as dict
    WSnm:   str     WSname;             like 'srim84Kr_Air_NTP'
    <Ret>   dict    {'Bm':'84Kr','A':84,'Sym':'Kr','Z':36,  
                     'Tg':'Air_NTP', 'WSnm':'srim84Kr_Air_NTP'}
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    if('_' not in WSnm):
        raise ValueError('WSnm invalid =',WSnm)
    s1,Tg = WSnm.split('_',1)       #->r#901) '_'分割は最初の１個のみ
    _,Bm = re.split(r'srim',s1)     #->r#902)
    _A   = re.findall('\d{1,3}',Bm) # 数字\d を 1~3文字繰返し{1,3}
    Sym  = re.sub('\d{1,3}','', Bm) # Aの部分を ''　で置き換え
    if(_A == []):
        raise ValueError('WSnm invalid =',WSnm,'Bm=',Bm)
    A = int(_A[0])
    Z = atomSym().Z(Sym)
    return({'Bm':Bm,'A':A,'Sym':Sym,'Z':Z,'Tg':Tg,'WSnm':WSnm})    

""" srimWS dump
----.---.---.---.---.---.---.---.---.---.---.---.---"""

def srimWSdump( ws, modL:list=[0], *, 
               intrvE:int=1, show=True):
    """ srimWS contents dump 
    ws:     srimWS      for dump
    <optional>
    modL:   [int,..]    dump mode list
        0: simple info. ; wsName,Beam,Target
        1: other  info. ; TrgDens, BraggC, SRIMver, Corded
        2: target compo.; TrgCmp table
        3: conv. factor ; LET unit conversion table
        4: table size   ; {min|max} of E, LETe,n,t, Rng, StLng, StLtr
        5: table contens; E, LET{e|n|t}, Rng, StLng, StLtr
        6: table size   ; LET{e|n|t}{L|H}
       -1: all of above 
    <optional: kwd args>
    intrvE: int         when mod=5, skip E interval 
    show:   bool        show dump result as print()
    <Ret>   [str,..]    dump result     
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    def _info0():   # Simple info.
        tx1 = ' wsName: {}, '.format(ws.wsName)
        tx2 = ' Beam: {}, Z:{:2d}, A:{:3d}'.format(
            atomSym().Asym(ws.IonA,ws.IonZ),ws.IonZ,ws.IonA)
        tx3 = ' Target: {}, TargetNm: {}'.format(
            ws.Target,ws.TargetNm)
        if(ws.Gas):
            tx3 += ' [Gas]'
        return( [tx1+tx2,tx3] )

    def _info1():   # Other info.
        msg =('TrgDens: {:.3e} [g/cm3], {:.3e} [atom/cm3], BraggC: {:.2%}'
              if(ws.Gas) else
              'TrgDens: {:.3f} [g/cm3], {:.3e} [atom/cm3], BraggC: {:.2%}')
        tx1 = msg.format(
            ws.TrgDens, ws.TrgDensAtm, ws.BraggC)
        tx3 ='SRIMver: {}, Corded: {}'.format(
            ws.SRIMver, ws.Corded)
        if(not ws.Gas):
            return( [tx1,tx3])
        tx2 = 'GasPtbl: {:.3e} [Pa] GasTtbl: {:.1f} [degC]'.format(
            ws.GasPtbl, ws.GasTtbl)
        return( [tx1,tx2,tx3] )

    def _info2():   # Target composition table
        hd1 = ' TrgCmp: Atom, Atom, Atomic,   Mass'
        hd2 = '         Name, Numb,    [%],    [%]'
        fmt = '           {:2s},  {:3d}, {:6.2f}, {:6.2f}'
        txL= [hd1,hd2]
        for i,dc in enumerate(ws._TrgCmp):
            vs = list(dc.values())
            tx = fmt.format( vs[0], vs[1], vs[2], vs[3] ) 
            txL.append( tx )
        return( txL )

    def _info3():   # LET unit convertion table
        hd1 = '    CnvF: Uid, factor,   unit'
        fmt = '            {:d}, {:7.2e}, {}'
        txL= [hd1]
        for i in range(len(srimWS_Uid)):
            tx = fmt.format( i, ws._CnvF[i], srimWS_Uid[i] )
            #print(i,tx)
            txL.append( tx )
        return( txL )

    def _info4():   # Table size
        tx1 = '      E: min {:.3e}, max {:.3e} [MeV/A]'.format(
            ws.minE, ws.maxE )
        tx2 = '   LETe: min {:.3e}, max {:.3e} [MeV/(mg/cm2)], '\
            'maxLETe2E {:.3e} [MeV/A]'.format(
            ws.minLETe, ws.maxLETe, ws.maxLETe2E )
        tx3 = '   LETn: min {:.3e}, max {:.3e} [MeV/(mg/cm2)], '\
            'maxLETn2E {:.3e} [MeV/A]'.format(
            ws.minLETn, ws.maxLETn, ws.maxLETn2E )
        tx4 = '   LETt: min {:.3e}, max {:.3e} [MeV/(mg/cm2)], '\
            'maxLETt2E {:.3e} [MeV/A]'.format(
            ws.minLETt, ws.maxLETt, ws.maxLETt2E )
        tx5 = '    Rng: min {:.3e}, max {:.3e} [um]'.format(
            ws.minRng, ws.maxRng)
        tx6 = '  StLng: min {:.3e}, max {:.3e} [um]'.format(
            ws.minStLng, ws.maxStLng)
        tx7 = '  StLtr: min {:.3e}, max {:.3e} [um]'.format(
            ws.minStLtr, ws.maxStLtr)
        return( [tx1,tx2,tx3,tx4,tx5,tx6,tx7] )

    def _info5():   # Table contents
        hd1 = '  E intrv={:2d}  LETe,       LETn,       LETt,       '\
              '   Rng,     StLng,   StLtr'
        hd2 = '  [MeV/A],    [MeV/(mg/cm2)],,,                   '\
              '   [um],    [um],    [um]'
        fmt1= '  {:10.4e}, {:10.4e}, {:10.4e}, {:10.4e}, {:7.2f}, {:7.2f}, {:7.2f}'
        fmt2= '  len= {:5d}, {:10d}, {:10d}, {:10d}, {:7d}, {:7d}, {:7d}'
        tx = hd1.format(intrvE)
        txL= [tx,hd2]
        for i,E in enumerate(ws._E):
            if((i % intrvE) == 0):
                tx = fmt1.format( E, ws._LETe[i], ws._LETn[i],  ws._LETt[i],
                                    ws._Rng[i],  ws._StLng[i], ws._StLtr[i])
                txL.append( tx )
        tx = fmt2.format(len(ws._E), 
                         len(ws._LETe), len(ws._LETn),  len(ws._LETt),
                         len(ws._Rng),  len(ws._StLng), len(ws._StLtr) )
        txL.append( tx )
        return( txL )

    def _info6():   # (E[] vs LETe[]) at Lower/Higher-side of its peak
        #hd1 = '     [MeV/A], [MeV/(mg/cm2)]'
        def _dmp(id1:str,id2:str, _LETlh, maxLET2E, maxLET):
            fmt1= ' _LET{}{}[][]     [MeV/A]      [MeV/(mg/cm2)] len={:3d}'
            fmt2= ' min [0][ 0] {:.4e}, [1][ 0] {:.4e} : maxLET{}2E   maxLET{}'
            fmt3= ' max [0][-1] {:.4e}, [1][-1] {:.4e} : {:.4e}, {:.4e}'
            tx1 = fmt1.format( id1, id2, len(_LETlh[0]) )
            tx2 = fmt2.format( _LETlh[0][0],  _LETlh[1][0], id1, id1 )
            tx3 = fmt3.format( _LETlh[0][-1], _LETlh[1][-1], maxLET2E, maxLET )
            return( [tx1,tx2,tx3] )

        txL = []
        txL += _dmp('e','L', ws._LETeL, ws.maxLETe2E, ws.maxLETe)
        txL += _dmp('e','H', ws._LETeH, ws.maxLETe2E, ws.maxLETe)
        txL += _dmp('n','L', ws._LETnL, ws.maxLETn2E, ws.maxLETn)
        txL += _dmp('n','H', ws._LETnH, ws.maxLETn2E, ws.maxLETn)
        txL += _dmp('t','L', ws._LETtL, ws.maxLETt2E, ws.maxLETt)
        txL += _dmp('t','H', ws._LETtH, ws.maxLETt2E, ws.maxLETt)
        return( txL )

    fnL = [_info0, _info1, _info2, _info3, _info4, _info5, _info6]
    txL = []
    for i in range(len(fnL)):
        if((i in modL) or (-1 in modL)):
            txL += fnL[i]() 
    if(show):       print('\n'.join(txL))
    return( txL )

""" srimWS plot
----.---.---.---.---.---.---.---.---.---.---.---.---"""

def _plot_fgax(rngX:tuple,rngY:tuple):
    fg = plt.figure(figsize=(7,7))
    #rngX,rngY = (1e-5,1e3),(1e-2,1e2)
    #rngX,rngY = (1e-3,1e1),(1e-1,1e1)
    ax = fg.add_subplot(1,1,1,xmargin=0,ymargin=0,xlim=rngX,ylim=rngY)
    return(fg,ax)

def _plot_plot(ax,X,Y,col,lab):
    ax.plot(X,Y,color=col)
    ax.scatter(X,Y,color=col,s=10,label=lab)

def _plot_ax_set(fg,ax,ttl,xlab,ylab,lgdfs=14):
    ax.set_title(ttl,fontsize=16)
    ax.set_xlabel(xlab,size=14)
    ax.set_ylabel(ylab,size=14)
    ax.set_xscale('log')
    ax.set_yscale('log')
    #xtick, ytick = ax.get_xticklabels(), ax.get_yticklabels()
    #ax.set_xticklabels(xtick,fontsize=14)
    #ax.set_yticklabels(ytick,fontsize=14)
    ax.tick_params(axis='x', labelsize=14)
    ax.tick_params(axis='y', labelsize=14)
    ax.grid(which='major',axis='both',lw=0.5,c='gainsboro', zorder=0)
    ax.grid(which='minor',axis='both',lw=0.1,c='gainsboro', zorder=0)
    ax.minorticks_on()
    ax.legend(fontsize=lgdfs)
    fg.tight_layout()

def srimWSplot_LET( ws, *, 
                LETe:bool=True, LETn:bool=True, LETt:bool=True,
                rngX:tuple=(1e-5,1e3), rngY=(1e-2,1e2) ) -> tuple:
    """ srimWS plot LET{e|n|t} 
    ws:     srimWS      for plot
    <optional: kwd args>
    LET{e|n|t}: bool    plot SW for LET{e|n|t}
    rngX,rngY:  tuple   plot range (min,max)
    <Ret> fg,ax         figure, axies
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    fg,ax = _plot_fgax(rngX,rngY)         
    if(LETt): _plot_plot(ax, ws._E, ws._LETt,col='g',lab='LETt')
    if(LETe): _plot_plot(ax, ws._E, ws._LETe,col='r',lab='LETe')
    if(LETn): _plot_plot(ax, ws._E, ws._LETn,col='b',lab='LETn')
    _plot_ax_set(fg,ax,ws.wsName,'E [MeV/A]','LET [MeV/(mg/cm2)]')
    return(fg,ax)

def srimWSplot_RngSt( ws, *, 
                Rng:bool=True, StLng:bool=True, StLtr:bool=True,
                rngX:tuple=(1e-5,1e3), rngY=(1e-3,1e6) ) -> tuple:
    """ srimWS plot Range, StLng, StLtr
    ws:     srimWS          for plot
    <optional: kwd args>
    Rng,StLng,StLtr: bool   plot SW for Rng,StLng,StLtr
    rngX,rngY:  tuple       plot range (min,max)
    <Ret> fg,ax             figure, axies
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    fg,ax = _plot_fgax(rngX,rngY)         
    if(Rng):    _plot_plot(ax, ws._E, ws._Rng,  col='r',lab='Rng')
    if(StLng):  _plot_plot(ax, ws._E, ws._StLng,col='b',lab='StLng')
    if(StLtr):  _plot_plot(ax, ws._E, ws._StLtr,col='g',lab='StLtr')
    _plot_ax_set(fg,ax,ws.wsName,'E [MeV/A]','Range, Straggling [um]')
    return(fg,ax)

def srimWSplot_LETcmp( wsL:list, *,
                LETe:bool=False, LETn:bool=False, LETt:bool=True,
                rngX:tuple=(1e-5,1e3), rngY=(1e-2,1e2) ) -> tuple:
    """ srimWS plot LET{e|n|t} comp. with plural srWS  
    wsL:      [srimWS]  for plot
    <optional: kwd args>
    LET{e|n|t}: bool    plot SW for LET{e|n|t}
    rngX,rngY:  tuple   plot range (min,max)
    <Ret> fg,ax         figure, axies
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    fg,ax = _plot_fgax(rngX,rngY)
    cmap = plt.get_cmap('tab10')
    ic=0
    for i,ws in enumerate(wsL):
        if(LETt):
            lb = 'LETt_'+ws.wsName
            _plot_plot(ax, ws._E, ws._LETt,col=cmap(ic),lab=lb)
            ic += 1
        if(LETe):
            lb = 'LETe_'+ws.wsName
            _plot_plot(ax, ws._E, ws._LETe,col=cmap(ic),lab=lb)
            ic += 1
        if(LETn): 
            lb = 'LETn_'+ws.wsName
            _plot_plot(ax, ws._E, ws._LETn,col=cmap(ic),lab=lb)
            ic += 1
    _plot_ax_set(fg,ax,'LET comp.',
                 'E [MeV/A]','LET [MeV/(mg/cm2)]',lgdfs=11)
    return(fg,ax)

def srimWSplot_RngStcmp( wsL:list, *,
                    Rng:bool=True, StLng:bool=False, StLtr:bool=False,
                    rngX:tuple=(1e-5,1e3), rngY=(1e-3,1e6) ) -> tuple:
    """ srimWS plot Range, StLng, StLtr comp. with plural srWS
    wsL:      [srimWS]      for plot
    <optional: kwd args>
    Rng,StLng,StLtr: bool   plot SW for Rng,StLng,StLtr
    rngX,rngY:  tuple       plot range (min,max)
    <Ret> fg,ax             figure, axies
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    fg,ax = _plot_fgax(rngX,rngY)
    cmap = plt.get_cmap('tab10')
    ic=0
    for i,ws in enumerate(wsL):
        if(Rng):
            lb = 'Rng_'+ws.wsName
            _plot_plot(ax, ws._E, ws._Rng,col=cmap(ic),lab=lb)
            ic += 1
        if(StLng): 
            lb = 'StLng_'+ws.wsName
            _plot_plot(ax, ws._E, ws._StLng,col=cmap(ic),lab=lb)
            ic += 1
        if(StLtr):
            lb = 'StLtr_'+ws.wsName
            _plot_plot(ax, ws._E, ws._StLtr,col=cmap(ic),lab=lb)
            ic += 1
    _plot_ax_set(fg,ax,'Range / Straggling comp.',
                 'E [MeV/A]','Range, Straggling [um]',lgdfs=11)
    return(fg,ax)


""" ========================================================
            Test Main
    ========================================================"""
if __name__ == '__main__':
    if(True):
        print('TEST : atomSym()')
        at = atomSym()
        print(at.sym(36))
        print(at.Asym(84,36))    
        print(at.Z('Kr'))
    if(True):
        pass
    
    print('@@main: -- Fin ---')


""" ========================================================
        ref) Numbered ref
    ========================================================
    
------------------------------------------------------------------
"""