# -*- coding: utf-8 -*-
"""
    Module: pySRIMfit
    File:   upd_MySRIMwb.py
    Descr.: Update MySRIMwb.xlsx
    Update: 230311 build start Ay
            230320 v.1.0 done
    Last:   230422 v.1.1 done
    
@author: Ayoshida(RIKEN,Nishina center)
"""
from    pySRIMfit   import  (srimWBupdater)

""" .===.===.===.===.===.===.===.===.===.===.===.===.===.===.===.===
            upd_MySRIMwb.py の 使い方
        ============================
[0] はじめに
============================================
  pySRIMfit の関数は、
    ~/pySRIMfit/_MySRIMwb/MySRIMwb.pickle 下図[1]で《重要》に置いてある
  を読み込んでから動作します。このファイルは、今迄 Excel-VBA 環境で使っていた
    MySRIMwb.xlsx ; MS-Excel版 SRIMfit 用の DB
  を、バイナリー形式(.pickle)　に変換したものです。

  このプログラム upd_MySRIMwb.py で、MySRIMwb.pickle　を編集します。
 　　・ 内容表示： シート名一覧
 　　・　SRIM-2013 で計算したファイルの追加
 　　・ シートの削除、追加、並べ替え
　　 ・ 編集結果の保存
  等の編集作業ができます。
   
[1] ファイル配置
============================================
DOS> tree /F ~/pyE5A
│  upd_MySRIMwb.py          <-- "This file"
│  MySRIMwb_tmp.pickle          <-- ~tmp.pickle
│  MySRIMwb_tmp.xlsx            <-- ~tmp.xlsx
│  MySRIMwb_add.xlsx            <-- ~add.xlsx
│  MySRIMwb_new.pickle          <-- ~new.pickle
│  MySRIMwb_new.xlsx            <-- ~new.xlsx
├─pySRIMfit
│  │  srimWBupdater.py          <-- program called by "This file"
│  │  ~ ~ ~ 
│  ├─_MySRIMwb              <-- Current DB folder 
│  │      MySRIMwb.pickle       <-- pySRIMfit DB  《重要》
│  │      MySRIMwb.xlsx         <-- same DB saved as .xlsx
│  ├─add_SRIMoutTXT         <-- SRIM-2013 data(.txt) to Add
│  │      Argon in Gold.txt     <-- sample data
│  │      ~ ~ ~
│  ├─add_SRIMoutCSV         <-- SRIM-2013 data(.xlsx) to Add
│  │      SRIMout_084Kr.xlsx    <-- sample data
│  │      ~ ~ ~

[2-1]　操作手順(1): 初めて pySRIMfit を使う　
============================================
 2-1a) 今迄 Excelで使っていた MySRIMwb.xlsx を。
        ~/_MySRIMwb フォルダー に 上書きコピー しておく
 2-1b) アップデータ:upd_MySRIMwb.py を起動すると、自動的に次のcopyが実行される
        ~/_MySRIMwb/ == オリジナル を   --> ./ == ワーク用 に コピー
        ~/_MySRIMwb/MySRIMwb.xlsx   --> ./MySRIMwb_tmp.xlsx
        ~/_MySRIMwb/MySRIMwb.pickle --> ./MySRIMwb_tmp.pickle
　2-1c) ワーク用としてコピーされた MySRIMwb_tmp.xlsx の 内容表示
    '==== MySRIMwb Updater ===='
    ' 1) Show    Contents'
    '---- Show Contents of MySRIMwb ----',
    ' 2) ~tmp.xlsx',
　2-1d) MySRIMwb_tmp.xlsx を MySRIMwb_new.pickle として保存
    '==== MySRIMwb Updater ====',
    ' 5) Save    Book',
    '---- Save ~tmp ----',
    ' 2) Save ~tmp.xlsx   as ~new.pickle & ~new.xlsx',
　2-1e) バイナリー変換された MySRIMwb_new.pickle を、
　　　　手動操作で、 ~/_MySRIMwb フォルダー に　
    　MySRIMwb.pickle　として　上書きコピー
 以上の操作で、今までの Excel環境と同じ DB を使って、pySRIMfit が動作します。
 
[2-2]　操作手順(2): SRIM-2013 の出力 (.txt) を追加する　
============================================
 2-2a) SRIM-2013 の出力 (*.txt) 複数ファイルも可 を。
        ~/add_SRIMoutTXT フォルダー に コピー しておく。不要な sampleは消しておく。
 2-2b) (*.txt) を　MySRIMwb_add.xlsx に変換する
    '==== MySRIMwb Updater ====',
    ' 3) Convert SRIM-2013 data',
    '---- Convert SRIM-2013 Output data to ~add.xlsx  ----',
    ' 1) Conv. SRIM TEXT to ~add.xlsx',
 2-2c) 変換結果　MySRIMwb_add.xlsx を MySRIMwb_tmp.pickle に追加する
    '==== MySRIMwb Updater ====',
    ' 4) Append  Sheet',
    '---- Append ~add.xlsx to ~tmp.pickle ----',
    ' 1) Check  sheet name list',
    ' 2) Append ~add.xlsx to ~tmp.pickle',
    メニュー 1) Check で、追加シートの重複が無いか？ を確認
    重複があった場合は、メインメニューに戻り、MySRIMwb_tmp.pickle から重複シートを削除
    '==== MySRIMwb Updater ====',
    ' 2) Remove  Sheet',
    重複がなくなったら、メニュー 2) Append で、追加する
　2-2d) MySRIMwb_tmp.xlsx を MySRIMwb_new.pickle として保存　
　　　2-1d) に同じ
　2-2e) 変換結果 MySRIMwb_new.pickle を　~/_MySRIMwb フォルダーに保存
　　　2-1e) に同じ

[2-3]　操作手順(3): SRIM-2013 の出力 (.txt) を 
      Excelで 手動操作で .csv として読み込んで　ブック形式にまとめて  
　　　　　*.xlsx として保存したファイルを 追加する
============================================
 2-3a) SRIM-2013 の出力 (*.xlsx) 複数ファイルも可 を。
        ~/add_SRIMoutCSV フォルダー に コピー しておく。不要な sampleは消しておく。
 2-3b) (*.xlsx) を　MySRIMwb_add.xlsx に変換する
    '==== MySRIMwb Updater ====',
    ' 3) Convert SRIM-2013 data',
    '---- Convert SRIM-2013 Output data to ~add.xlsx  ----',
    ' 2) Conv. SRIM CSV  to ~add.xlsx',
 2-3c) 変換結果　MySRIMwb_add.xlsx を MySRIMwb_tmp.pickle に追加する
    2-2c) に同じ
　2-2d) MySRIMwb_tmp.xlsx を MySRIMwb_new.pickle として保存　
　　　2-2d) に同じ
　2-2e) 変換結果 MySRIMwb_new.pickle を　~/_MySRIMwb フォルダーに保存
　　　2-2e) に同じ

以上です。
====.===.===.===.===.===.===.===.===.===.===.===.===.===.===.=== === """

""" .===.===.===.===.===.===.===.===.===.===.===.===.===.===.===.===
            Main
    note) 
    もし途中終了した場合は、Xlwings が開いた タスクが残る場合があります。
    その場合は、タスクマネージャーから、Microsoft Excel を「タスク終了」してください。
====.===.===.===.===.===.===.===.===.===.===.===.===.===.===.=== === """
if __name__ == '__main__':
    srimWBupdater()
    print('@@main: -- Fin ---')

